﻿// ***********************************************************************************
// SallySoundFX
// ***********************************************************************************
// Created by Sally Whitemane on VAM Hub
// https://www.patreon.com/sallyw
// Licensed under CC BY-SA - https://creativecommons.org/licenses/by-sa/4.0/

using System;
using UnityEngine;
using UnityEngine.UI;

namespace Sally
{
    public class AVSLAVE_Graph : MVRScript
    {
        // UI
        private JSONStorableString jsBenchmark;
        private JSONStorableBool jbSyncTrigger1;
        private JSONStorableBool jbSyncTrigger2;
        private JSONStorableFloat jfFrequency;
        private UIDynamicSlider sliderFrequency;
        private JSONStorableFloat jfSampleValue;
        private JSONStorableFloat jfFramerateLimit;
        private JSONStorableFloat jfCubeScale;

        // BENCHMARK keep namespace, can cause VAM compilation error
        private System.Diagnostics.Stopwatch stopWatch = new System.Diagnostics.Stopwatch();
        private double[] avgUpdateTimes = new double[80];
        private double avgUpdateTime = 0;

        // AUDIO VISUALIZATION DATA
        private AVD avd = new AVD();

        // VISUALIZATION GRAPH
        private GameObject[] cubes = new GameObject[AVD.SAMPLES];
        private MeshRenderer[] meshRenderers = new MeshRenderer[AVD.SAMPLES];
        private Material cubeMaterial;

        // MISCELLANEOUS
        private float cubeYMaxPosition = 0.5f;
        private Vector3 newPosition; // optimization
        private Vector3 cubeScaleV3;  // optimization
        private int sampleIndexFrequency = 2; // index for avd.samples[], restored by SetFrequency() on Scene load
        private float deltaTimeCounter = 0f; // used to limit Visualization framerate

        public override void Init()
        {
            try
            {
                if (AVSlave.GetMasterScript(this) == null)
                    SuperController.LogMessage($"{name} could not find SallySoundFX Plugin to synchronize. It will retry automatically.");

                //////////////////////////////////////////////////////////////////////////////////
                // UI LEFT
                //////////////////////////////////////////////////////////////////////////////////

                // Benchmark
                jsBenchmark = new JSONStorableString("Benchmark", "no data");
                UIDynamicTextField tf = CreateTextField(jsBenchmark, false);
                LayoutElement le = tf.GetComponent<LayoutElement>();
                le.minHeight = 40.0f;
                le.preferredHeight = 40.0f;

                // Toggle Sync Trigger 1
                jbSyncTrigger1 = new JSONStorableBool("Sync to SSFX Trigger 1", true);
                RegisterBool(jbSyncTrigger1);
                CreateToggle(jbSyncTrigger1, false).labelText.color = AVD.trigger1Color;
                jbSyncTrigger1.setCallbackFunction += SetSyncTrigger1;

                // Toggle Sync Trigger 2
                jbSyncTrigger2 = new JSONStorableBool("Sync to SSFX Trigger 2", false);
                RegisterBool(jbSyncTrigger2);
                CreateToggle(jbSyncTrigger2, false).labelText.color = AVD.trigger2Color;
                jbSyncTrigger2.setCallbackFunction += SetSyncTrigger2;

                // Slider to select Frequency
                jfFrequency = new JSONStorableFloat("Frequency in Hz", AVD.SampleFreq(), AVD.SampleFreq(), AVD.SampleFreq() * AVD.SAMPLES);
                RegisterFloat(jfFrequency);
                jfFrequency.setCallbackFunction += SetFrequency;
                sliderFrequency = CreateSlider(jfFrequency);
                sliderFrequency.Configure(jfFrequency.name, jfFrequency.min, jfFrequency.max, AVD.SampleFreq() * (sampleIndexFrequency + 1), true, "F3", false, false);
                sliderFrequency.labelText.text = $"Frequency {jfFrequency.val} Hz";
                le = sliderFrequency.GetComponent<LayoutElement>();
                le.minWidth = 1060;

                // Slider to show Value [no register]
                jfSampleValue = new JSONStorableFloat("Sample Value", 0.0f, 0.0f, 1.0f);
                UIDynamicSlider sliderValue = CreateSlider(jfSampleValue);
                sliderValue.quickButtonsEnabled = false;
                sliderValue.defaultButtonEnabled = false;
                sliderValue.GetComponentInChildren<Image>().color = UI.blueAutoBackground;
                le = sliderValue.GetComponent<LayoutElement>();
                le.minWidth = 1060;

                // Slider to limit Framerate
                jfFramerateLimit = new JSONStorableFloat("Framerate Limit", 60.0f, 1.0f, 240.0f);
                RegisterFloat(jfFramerateLimit);
                UIDynamicSlider slider = CreateSlider(jfFramerateLimit);
                slider.quickButtonsEnabled = false;
                le = slider.GetComponent<LayoutElement>();
                le.minWidth = 1060;

                //////////////////////////////////////////////////////////////////////////////////
                // UI RIGHT
                //////////////////////////////////////////////////////////////////////////////////

                // Slider Cube Scale
                jfCubeScale = new JSONStorableFloat("Cube Scale", 0.01f, 0.001f, 0.05f);
                RegisterFloat(jfCubeScale);
                slider = CreateSlider(jfCubeScale, true);
                slider.GetComponentInChildren<Image>().color = UI.redWarningBackground;
                jfCubeScale.setCallbackFunction += RebuildGraphFloat;

                //////////////////////////////////////////////////////////////////////////////////
                // VISUALIZATION
                //////////////////////////////////////////////////////////////////////////////////

                CreateVisualization(true);
            }
            catch (Exception e) { SuperController.LogError("Exception: " + e); }
        }

        private void CreateVisualization(bool initialize = false)
        {
            // Atom rotation workaround - save original state
            // makes the Visualization stay in sync with Atom-rotation after scene load
            Quaternion rotationOriginal = containingAtom.mainController.transform.rotation;
            Quaternion localRotationOriginal = containingAtom.mainController.transform.localRotation;
            // reset rotation
            containingAtom.mainController.transform.rotation = Quaternion.identity;
            containingAtom.mainController.transform.localRotation = Quaternion.identity;

            if (initialize)
            {
                cubeMaterial = new Material(Shader.Find("Sprites/Default")); // Sprites/Default  -  Custom/Subsurface/EmissiveGlow
                cubeScaleV3 = new Vector3(jfCubeScale.val, jfCubeScale.val, jfCubeScale.val);
            }

            cubeScaleV3.x = jfCubeScale.val;
            cubeScaleV3.y = jfCubeScale.val;
            cubeScaleV3.z = jfCubeScale.val;
            float x = 0.0f;
            float xoffset = jfCubeScale.val; // + 0.01f;
            Vector3 xoffsetV3 = new Vector3(0f, 0f, 0f);
            for (int i = 0; i < cubes.Length; i++)
            {
                if (initialize)
                {
                    cubes[i] = GameObject.CreatePrimitive(PrimitiveType.Cube);
                    cubes[i].transform.parent = containingAtom.mainController.transform;
                }
                xoffsetV3.x = -x;
                cubes[i].transform.position = containingAtom.mainController.transform.position + xoffsetV3;//new Vector3(-x, 0f, 0f);
                cubes[i].transform.rotation = containingAtom.mainController.transform.rotation;
                cubes[i].transform.localScale = cubeScaleV3;
                meshRenderers[i] = cubes[i].GetComponent<MeshRenderer>();
                meshRenderers[i].material = cubeMaterial;
                x += xoffset;

                // remove Box Collider
                UnityEngine.Object.DestroyImmediate(cubes[i].GetComponent<Collider>());
            }

            // Atom rotation workaround - restore original state
            containingAtom.mainController.transform.rotation = rotationOriginal;
            containingAtom.mainController.transform.localRotation = localRotationOriginal;
        }

        void Update()
        {
            try
            {
                // BENCHMARK START
                stopWatch.Reset();
                stopWatch.Start();

                if (!AVSlave.IsSynchronized(this))
                    return;

                // Sync Trigger 1 ?
                if (jbSyncTrigger1.val)
                    // Update Trigger 1 Sync
                    jfFrequency.val = (avd.indexTrigger1 + 1) * AVD.SampleFreq();
                // Sync Trigger 2 ?
                else if (jbSyncTrigger2.val)
                    // Update Trigger 1 Sync
                    jfFrequency.val = (avd.indexTrigger2 + 1) * AVD.SampleFreq();

                // Update Value Slider
                jfSampleValue.val = avd.samples[sampleIndexFrequency];

                // framerate limit
                deltaTimeCounter += Time.deltaTime;
                if (deltaTimeCounter > 1f / jfFramerateLimit.val)
                {
                    deltaTimeCounter = deltaTimeCounter - 1f / jfFramerateLimit.val;

                    // shift y position, but keep x and z
                    for (int i = AVD.SAMPLES - 1; i > 0; i--)
                    {
                        newPosition = meshRenderers[i].transform.position;
                        newPosition.y = meshRenderers[i - 1].transform.position.y;
                        meshRenderers[i].transform.position = newPosition;
                    }
                    // update the first cube with current frequency value
                    newPosition = meshRenderers[0].transform.position;
                    newPosition.y = avd.samples[sampleIndexFrequency] * cubeYMaxPosition;
                    meshRenderers[0].transform.position = containingAtom.mainController.transform.position + newPosition;
                }

                // BENCHMARK END
                stopWatch.Stop();
                avgUpdateTimes[avgUpdateTimes.Length - 1] = stopWatch.Elapsed.TotalMilliseconds;
                Array.Copy(avgUpdateTimes, 1, avgUpdateTimes, 0, avgUpdateTimes.Length - 1); // shift array
                avgUpdateTime = 0;
                for (int i = 0; i < avgUpdateTimes.Length; i++)
                    avgUpdateTime += avgUpdateTimes[i];
                avgUpdateTime = avgUpdateTime / avgUpdateTimes.Length;

                jsBenchmark.val = $"Update Benchmark: {avgUpdateTime:F4} ms";
            }
            catch (Exception e) { SuperController.LogError("Exception: " + e); }
        }

        void OnDestroy()
        {
            int i = 0;

            // cubes
            for (i = 0; i < cubes.Length; i++)
            {
                Destroy(cubes[i]);
                cubes[i] = null;
            }
            cubes = null;

            // MeshRenderer
            for (i = 0; i < meshRenderers.Length; i++)
            {
                Destroy(meshRenderers[i]);
                meshRenderers[i] = null;
            }
            meshRenderers = null;

            // cubeMaterial
            if (cubeMaterial != null)
            {
                Destroy(cubeMaterial);
                cubeMaterial = null;
            }

            avgUpdateTimes = null;
            avd = null;
        }

        //////////////////////////////////////////////////////////////////////////////////
        // UI Callbacks
        //////////////////////////////////////////////////////////////////////////////////
        private void SetFrequency(float f)
        {
            f = AVD.SnapToSampleFrequency(f);
            jfFrequency.val = f;

            if (jbSyncTrigger1.val)
                sliderFrequency.labelText.color = AVD.trigger1Color;
            else if (jbSyncTrigger2.val)
                sliderFrequency.labelText.color = AVD.trigger2Color;
            else
                sliderFrequency.labelText.color = Color.black;

            sliderFrequency.labelText.text = $"Frequency {jfFrequency.val} Hz";
            sampleIndexFrequency = (int)(f / AVD.SampleFreq()) - 1;
        }

        private void SetSyncTrigger1(bool b)
        {
            if (b) jbSyncTrigger2.val = false;
            if (jbSyncTrigger1.val)
                sliderFrequency.labelText.color = AVD.trigger1Color;
            else if (!jbSyncTrigger1.val && !jbSyncTrigger2.val)
                sliderFrequency.labelText.color = Color.black;
        }
        private void SetSyncTrigger2(bool b)
        {
            if (b) jbSyncTrigger1.val = false;
            if (jbSyncTrigger2.val)
                sliderFrequency.labelText.color = AVD.trigger2Color;
            else if (!jbSyncTrigger1.val && !jbSyncTrigger2.val)
                sliderFrequency.labelText.color = Color.black;
        }

        private void RebuildGraphFloat(float v) { CreateVisualization(); }

        //////////////////////////////////////////////////////////////////////////////////
        // AVD Synchronization - called by SendMessage from Master
        //////////////////////////////////////////////////////////////////////////////////
        public void Sync(float[] syncdata)
        {
            try
            {
                avd.UnpackSyncdata(syncdata);
            }
            catch (Exception e) { SuperController.LogError("Exception: " + e); }
        }
    }
}
