﻿// ***********************************************************************************
// SallySoundFX
// ***********************************************************************************
// Created by Sally Whitemane on VAM Hub
// https://www.patreon.com/sallyw
// Licensed under CC BY-SA - https://creativecommons.org/licenses/by-sa/4.0/

using System;
using UnityEngine;
using UnityEngine.UI;

namespace Sally
{
    public class AVSLAVE_Grid : MVRScript
	{
		// UI - LEFT
		private JSONStorableColor jColorHi;
		private JSONStorableColor jColorLo;

		// UI - RIGHT
		private JSONStorableString jsBenchmark;
		private JSONStorableBool jbShowTriggerFreq;
		private JSONStorableBool jbUsePeaks;
		private JSONStorableFloat jfScale;
		private JSONStorableFloat jfHeight;
		private JSONStorableFloat jfFramerateLimit;

		// BENCHMARK keep namespace, can cause VAM compilation error
		private System.Diagnostics.Stopwatch stopWatch = new System.Diagnostics.Stopwatch();
		private double[] avgUpdateTimes = new double[80];
		private double avgUpdateTime = 0;

		// AUDIO VISUALIZATION DATA
		private AVD avd = new AVD();

		// VISUALIZATION
		private readonly int xSize = AVD.SAMPLES - 1; // size in quads
		private readonly int zSize = AVD.SAMPLES - 1; // size in quads
		private GameObject goGrid;
		private float height;
		private float[] source; // signal source, set to avd.samples or samples_peak
        private float deltaTimeCounter = 0f; // used to limit Visualization framerate
		// VISUALIZATION MESH
		private Mesh mesh;
		private Material material;
		private Color defaultColor = Color.white;
		private Vector3[] vertices;
		private Color[] colors;
		private Color colorLoRGB = Color.cyan;
		private Color colorHiRGB = Color.magenta;

		// OPTIMIZATION
		private MeshFilter meshFilter;
		private const int SAMPLES = AVD.SAMPLES;
		//private int indexMinusXSizeVert;
		//private int indexEndArrayShift = AVD.SAMPLES * 2;

		public override void Init()
		{
			try
			{
				if (AVSlave.GetMasterScript(this) == null)
					SuperController.LogMessage($"{name} could not find SallySoundFX Plugin to synchronize. It will retry automatically.");

				source = avd.samples;

				//////////////////////////////////////////////////////////////////////////////////
				// UI - LEFT
				//////////////////////////////////////////////////////////////////////////////////
				UIDynamicSlider slider;

				// Color Picker High
				jColorHi = new JSONStorableColor("Color High", HSVColorPicker.RGBToHSV(1f, 0f, 1f), SetHiColor);
				RegisterColor(jColorHi);
				CreateColorPicker(jColorHi);
				UI.FixInitialColorPickerSelectorPosition(jColorHi);

				// Color Picker Low
				jColorLo = new JSONStorableColor("Color Low", HSVColorPicker.RGBToHSV(0f, 1f, 1f), SetLoColor);
				RegisterColor(jColorLo);
				CreateColorPicker(jColorLo);
				UI.FixInitialColorPickerSelectorPosition(jColorLo);

				//////////////////////////////////////////////////////////////////////////////////
				// UI - RIGHT
				//////////////////////////////////////////////////////////////////////////////////

				// Benchmark
				jsBenchmark = new JSONStorableString("Benchmark", "no data");
				UIDynamicTextField tf = CreateTextField(jsBenchmark, true);
				LayoutElement le = tf.GetComponent<LayoutElement>();
				le.minHeight = 40.0f;
				le.preferredHeight = 40.0f;

				// Show Trigger Frequencies
				jbShowTriggerFreq = new JSONStorableBool("Show Trigger Frequencies", false);
				RegisterBool(jbShowTriggerFreq);
				CreateToggle(jbShowTriggerFreq, true);

				// Toggle Use Peak Samples
				jbUsePeaks = new JSONStorableBool("Use Peaks", false);
				RegisterBool(jbUsePeaks);
				CreateToggle(jbUsePeaks, true);
				jbUsePeaks.setCallbackFunction += SetUsePeakSamples;

				// Slider Scale
				jfScale = new JSONStorableFloat("Scale", 0.01f, 0.001f, 0.1f);
				RegisterFloat(jfScale);
				slider = CreateSlider(jfScale, true);
				slider.Configure(jfScale.name, jfScale.min, jfScale.max, jfScale.val, true, "F4");
				jfScale.setCallbackFunction += SetScale;

				// Slider Height
				jfHeight = new JSONStorableFloat("Height", 5f, 0f, 50f);
				RegisterFloat(jfHeight);
				slider = CreateSlider(jfHeight, true);
				jfHeight.setCallbackFunction += SetHeight;

				// Slider to limit Framerate
				jfFramerateLimit = new JSONStorableFloat("Framerate Limit", 60.0f, 1.0f, 240.0f);
				RegisterFloat(jfFramerateLimit);
				 CreateSlider(jfFramerateLimit, true);

				//////////////////////////////////////////////////////////////////////////////////
				// VISUALIZATION
				//////////////////////////////////////////////////////////////////////////////////

				CreateVisualization(true);
				SetHeight(jfHeight.val);
			}
			catch (Exception e) { SuperController.LogError("Exception: " + e); }
		}

		private void CreateVisualization(bool initialize = false)
		{
			if (initialize)
            {
				goGrid = new GameObject("goGrid", typeof(MeshFilter), typeof(MeshRenderer));
				goGrid.GetComponent<MeshFilter>().mesh = mesh = new Mesh();
				mesh.name = "Procedural Grid";
				
				vertices = new Vector3[(xSize + 1) * (zSize + 1)];
				colors = new Color[vertices.Length];
				Vector2[] uv = new Vector2[vertices.Length];
				//Vector4[] tangents = new Vector4[vertices.Length];
				//Vector4 tangent = new Vector4(1f, 0f, 0f, -1f);
				for (int i = 0, z = 0; z <= zSize; z++)
				{
					for (int x = 0; x <= xSize; x++, i++)
					{
						vertices[i] = new Vector3(x, 0, z);
						colors[i] = defaultColor;
						uv[i] = new Vector2((float)x / xSize, (float)z / zSize);
						//tangents[i] = tangent;
					}
				}
				mesh.vertices = vertices;
				mesh.uv = uv;
				//mesh.tangents = tangents;
				mesh.colors = colors;

				int[] triangles = new int[xSize * zSize * 6];
				for (int ti = 0, vi = 0, z = 0; z < zSize; z++, vi++)
				{
					for (int x = 0; x < xSize; x++, ti += 6, vi++)
					{
						triangles[ti] = vi;
						triangles[ti + 3] = triangles[ti + 2] = vi + 1;
						triangles[ti + 4] = triangles[ti + 1] = vi + xSize + 1;
						triangles[ti + 5] = vi + xSize + 2;
					}
				}
				mesh.triangles = triangles;
				//mesh.RecalculateNormals();

				goGrid.GetComponent<MeshRenderer>().material = material = new Material(Shader.Find("UI/Default"));

				goGrid.transform.parent = containingAtom.mainController.transform;
				SetScale(jfScale.val);
				meshFilter = goGrid.GetComponent<MeshFilter>();

				//Debug.Log($"meshFilter.mesh.colors.Length: {meshFilter.mesh.colors.Length}");
			}
		}

		// Start is called once before Update or FixedUpdate is called and after Init()
		/*
		void Start()
		{
			try
			{
			}
			catch (Exception e) { SuperController.LogError("Exception: " + e); }
		}
		*/

		// Update is called with each rendered frame by Unity
		void Update()
		{
			try
			{
				// BENCHMARK START
				stopWatch.Reset();
				stopWatch.Start();

				if (!AVSlave.IsSynchronized(this))
					return;

				// framerate limit
				deltaTimeCounter += Time.deltaTime;
				if (deltaTimeCounter > 1f / jfFramerateLimit.val)
				{
					deltaTimeCounter = deltaTimeCounter - 1f / jfFramerateLimit.val;

					// modify mesh
					vertices = meshFilter.mesh.vertices;
					colors = meshFilter.mesh.colors;

					// shift shamples
					for (int i = vertices.Length - 1; i >= SAMPLES; i--)
					{
						vertices[i].y = vertices[i - SAMPLES].y;
						colors[i] = colors[i - SAMPLES];
					}

					// current samples
					float v;
					for (int i = 0; i <= xSize; i++)
					{
						v = source[i];
						vertices[i].y = v * height;
						// set vertex colors
						colors[i] = colorLoRGB * (1f - v) + v * colorHiRGB;
						if (jbShowTriggerFreq.val)
						{
							if (i == avd.indexTrigger1)
								colors[i] = AVD.trigger1Color;
							if (i == avd.indexTrigger2)
								colors[i] = AVD.trigger2Color;
						}
					}
					meshFilter.mesh.vertices = vertices;
					meshFilter.mesh.colors = colors;
					//mesh.RecalculateNormals();


				}

				// BENCHMARK END
				stopWatch.Stop();
				avgUpdateTimes[avgUpdateTimes.Length - 1] = stopWatch.Elapsed.TotalMilliseconds;
				Array.Copy(avgUpdateTimes, 1, avgUpdateTimes, 0, avgUpdateTimes.Length - 1); // shift array
				avgUpdateTime = 0;
				for (int i = 0; i < avgUpdateTimes.Length; i++)
					avgUpdateTime += avgUpdateTimes[i];
				avgUpdateTime = avgUpdateTime / avgUpdateTimes.Length;

				jsBenchmark.val = $"Update Benchmark: {avgUpdateTime:F4} ms";
			}
			catch (Exception e) { SuperController.LogError("Exception: " + e); }
		}

		// OnDestroy is where you should put any cleanup
		// if you registered objects to supercontroller or atom, you should unregister them here
		void OnDestroy()
		{
			try
			{
				avgUpdateTimes = null;
				avd = null;
				source = null;

				if (goGrid)
                {
					Destroy(goGrid);
					goGrid = null;
				}
				if (mesh != null)
                {
					mesh.Clear();
					Destroy(mesh);
					mesh = null;
                }
				if (material != null)
                {
					Destroy(material);
					material = null;
				}
				vertices = null;
				colors = null;
	}
			catch (Exception e) { SuperController.LogError("Exception: " + e); }
		}

		//////////////////////////////////////////////////////////////////////////////////
		// UI Callbacks
		//////////////////////////////////////////////////////////////////////////////////
		private void SetHiColor(float h, float s, float v) { colorHiRGB = Color.HSVToRGB(h, s, v); }
		private void SetLoColor(float h, float s, float v) { colorLoRGB = Color.HSVToRGB(h, s, v); }
		private void SetScale(float v)
		{
			//Debug.Log($"{name} SetScale() {v}");
			if (goGrid != null)
			{
				goGrid.transform.localScale = new Vector3(v, v, v);
				SyncToAtom();
			}
		}

		private void SetHeight(float h)
        {
			//Debug.Log($"{name} SetHeight() {h}");
			height = h;
        }

		private void SetUsePeakSamples(bool b)
        {
			if (b)
				source = avd.samples_peak;
			else
				source = avd.samples;
        }

		//////////////////////////////////////////////////////////////////////////////////
		// Helper functions
		//////////////////////////////////////////////////////////////////////////////////

		private void SyncToAtom()
		{
			// move goTiles to atom
			goGrid.transform.position = containingAtom.mainController.transform.position;
			goGrid.transform.rotation = containingAtom.mainController.transform.rotation;
			// center goTiles on atom
			goGrid.transform.Translate(new Vector3(-xSize * 0.5f * jfScale.val, 0.0f, -zSize * 0.5f * jfScale.val));
		}

		//////////////////////////////////////////////////////////////////////////////////
		// AVD Synchronization - called by SendMessage from Master
		//////////////////////////////////////////////////////////////////////////////////
		public void Sync(float[] syncdata)
		{
			try
			{
				avd.UnpackSyncdata(syncdata);
			}
			catch (Exception e) { SuperController.LogError("Exception: " + e);}
		}
	}
}
