﻿// ***********************************************************************************
// SallySoundFX
// ***********************************************************************************
// Created by Sally Whitemane on VAM Hub
// https://www.patreon.com/sallyw
// Licensed under CC BY-SA - https://creativecommons.org/licenses/by-sa/4.0/

using System;
using UnityEngine;
using UnityEngine.UI;

namespace Sally
{
    public class AVSLAVE_Spectrum : MVRScript
    {
        // UI LEFT
        private JSONStorableColor jColorHi;
        private JSONStorableColor jColorLo;
        // UI RIGHT
        private JSONStorableString jsBenchmark;
        private JSONStorableBool jbShowTriggerFreq;
        private JSONStorableFloat jfCubeScale;
        private JSONStorableFloat jfCubeXOffset;
        private JSONStorableFloat jfHeightScale;
        private JSONStorableFloat jfLimitSamples;
        
        // BENCHMARK keep namespace, can cause VAM compilation error
        private System.Diagnostics.Stopwatch stopWatch = new System.Diagnostics.Stopwatch();
        private double[] avgUpdateTimes = new double[80];
        private double avgUpdateTime = 0;

        // AUDIO VISUALIZATION DATA
        private AVD avd = new AVD();

        // VISUALIZATION SPECTRUM
        private MeshRenderer[] meshRenderers = new MeshRenderer[AVD.SAMPLES];
        private MeshFilter[] meshFilters = new MeshFilter[AVD.SAMPLES];
        private GameObject[] cubes = new GameObject[AVD.SAMPLES];
        private Material cubeMaterial;

        // MISCELLANEOUS
        Vector3 newScale;
        Color newColor = Color.white;
        Color[] colors = new Color[24]; // colors.Length must be mesh.vertices.Length!
        Color colorLoRGB = Color.cyan;
        Color colorHiRGB = Color.magenta;

        public override void Init()
        {
            try
            {
                if (AVSlave.GetMasterScript(this) == null)
                    SuperController.LogMessage($"{name} could not find SallySoundFX Plugin to synchronize. It will retry automatically.");

                //////////////////////////////////////////////////////////////////////////////////
                // UI LEFT
                //////////////////////////////////////////////////////////////////////////////////

                // Color Picker High
                jColorHi = new JSONStorableColor("Color High", HSVColorPicker.RGBToHSV(1f, 0f, 1f), SetHiColor);
                RegisterColor(jColorHi);
                CreateColorPicker(jColorHi);
                UI.FixInitialColorPickerSelectorPosition(jColorHi);

                // Color Picker Low
                jColorLo = new JSONStorableColor("Color Low", HSVColorPicker.RGBToHSV(0f, 1f, 1f), SetLoColor);
                RegisterColor(jColorLo);
                CreateColorPicker(jColorLo);
                UI.FixInitialColorPickerSelectorPosition(jColorLo);

                //////////////////////////////////////////////////////////////////////////////////
                // UI RIGHT
                //////////////////////////////////////////////////////////////////////////////////

                // Benchmark
                jsBenchmark = new JSONStorableString("Benchmark", "no data");
                UIDynamicTextField tf = CreateTextField(jsBenchmark, true);
                LayoutElement le = tf.GetComponent<LayoutElement>();
                le.minHeight = 40.0f;
                le.preferredHeight = 40.0f;

                // Show Trigger Frequencies
                jbShowTriggerFreq = new JSONStorableBool("Show Trigger Frequencies", false);
                RegisterBool(jbShowTriggerFreq);
                CreateToggle(jbShowTriggerFreq, true);

                // Slider Cube Scale
                jfCubeScale = new JSONStorableFloat("Cube Scale", 0.01f, 0.001f, 0.1f);
                RegisterFloat(jfCubeScale);
                UIDynamicSlider sliderCubeScale = CreateSlider(jfCubeScale, true);
                sliderCubeScale.Configure(jfCubeScale.name, jfCubeScale.min, jfCubeScale.max, jfCubeScale.val, jfCubeScale.constrained, "F4", true, false);
                sliderCubeScale.GetComponentInChildren<Image>().color = UI.redWarningBackground;
                jfCubeScale.setCallbackFunction += RebuildSpectrum;

                // Slider Cube X Offset
                jfCubeXOffset = new JSONStorableFloat("Cube X Offset", 0.01f, 0.0f, 0.1f);
                RegisterFloat(jfCubeXOffset);
                UIDynamicSlider sliderCubeXOffset = CreateSlider(jfCubeXOffset, true);
                sliderCubeXOffset.Configure(jfCubeXOffset.name, jfCubeXOffset.min, jfCubeXOffset.max, jfCubeXOffset.val, jfCubeXOffset.constrained, "F4", true, false);
                sliderCubeXOffset.GetComponentInChildren<Image>().color = UI.redWarningBackground;
                jfCubeXOffset.setCallbackFunction += RebuildSpectrum;

                // Slider Height Scale
                jfHeightScale = new JSONStorableFloat("Height Scale", 1.0f, 0.1f, 5.0f);
                RegisterFloat(jfHeightScale);
                CreateSlider(jfHeightScale, true);

                // Slider Limit Samples
                jfLimitSamples = new JSONStorableFloat("Limit Samples", (float)AVD.SAMPLES, 1.0f, (float)AVD.SAMPLES);
                RegisterFloat(jfLimitSamples);
                CreateSlider(jfLimitSamples, true).Configure(jfLimitSamples.name, jfLimitSamples.min, jfLimitSamples.max, jfLimitSamples.val, jfLimitSamples.constrained, "F0");
                jfLimitSamples.setCallbackFunction += SetLimitSamples;

                //////////////////////////////////////////////////////////////////////////////////
                // VISUALIZATION
                //////////////////////////////////////////////////////////////////////////////////

                CreateVisualization(true);
            }
            catch (Exception e) { SuperController.LogError("Exception: " + e); }
        }

        private void CreateVisualization(bool initialize = false)
        {
            // Atom rotation workaround - save original state
            // makes the Visualization stay in sync with Atom-rotation after scene load
            Quaternion rotationOriginal = containingAtom.mainController.transform.rotation;
            Quaternion localRotationOriginal = containingAtom.mainController.transform.localRotation;
            // reset rotation
            containingAtom.mainController.transform.rotation = Quaternion.identity;
            containingAtom.mainController.transform.localRotation = Quaternion.identity;

            // CUBES
            if (initialize)
            {
                cubeMaterial = new Material(Shader.Find("Sprites/Default")); // Sprites/Default  -  Custom/Subsurface/EmissiveGlow
                cubeMaterial.color = Color.white;
            }
            newScale = new Vector3(jfCubeScale.val, jfCubeScale.val, jfCubeScale.val);

            float x = 0.0f;
            float xoffset = jfCubeScale.val + jfCubeXOffset.val;
            Vector3 initialScale = new Vector3(jfCubeScale.val, jfCubeScale.val, jfCubeScale.val);
            for (int i = 0; i < cubes.Length; i++)
            {
                if (initialize)
                    cubes[i] = GameObject.CreatePrimitive(PrimitiveType.Cube);
                cubes[i].transform.parent = containingAtom.mainController.transform;
                cubes[i].transform.position = containingAtom.mainController.transform.position + new Vector3(-x, 0.0f, 0.0f);
                cubes[i].transform.localScale = initialScale;
                //cubes[i].transform.name = "cube" + i;
                meshRenderers[i] = cubes[i].GetComponent<MeshRenderer>();
                meshRenderers[i].material = cubeMaterial;
                meshFilters[i] = cubes[i].GetComponent<MeshFilter>();
                x += xoffset;

                // remove Box Collider
                UnityEngine.Object.DestroyImmediate(cubes[i].GetComponent<Collider>());
            }

            // Atom rotation workaround - restore original state
            containingAtom.mainController.transform.rotation = rotationOriginal;
            containingAtom.mainController.transform.localRotation = localRotationOriginal;
        }

        void Update()
        {
            try
            {
                // DEBUG
                //if (Input.GetKey("left ctrl") || Input.GetKey("right ctrl"))
                //    if (Input.GetKeyDown("d"))
                //        myDebug();

                // BENCHMARK START
                stopWatch.Reset();
                stopWatch.Start();

                if (!AVSlave.IsSynchronized(this))
                    return;

                int c; // optimization
                float v; // optimization
                float v_inverted; // optimization
                //for (int i = 0; i < AVD.SAMPLES; i++)
                for (int i = 0; i < (int)jfLimitSamples.val; i++)
                {
                    v = avd.samples[i];
                    v_inverted = 1f - v;
                    newScale.y = jfCubeScale.val + jfHeightScale.val * v;
                    newColor.r = colorLoRGB.r * v_inverted + colorHiRGB.r * v;
                    newColor.g = colorLoRGB.g * v_inverted + colorHiRGB.g * v;
                    newColor.b = colorLoRGB.b * v_inverted + colorHiRGB.b * v;
                    //meshRenderers[i].transform.rotation = containingAtom.transform.rotation;
                    meshRenderers[i].transform.localScale = newScale;
                    // set material color
                    //meshRenderers[i].material.color = newColor; // 0.120 ms per loop

                    // set vertex colors
                    if (jbShowTriggerFreq.val)
                    {
                        if (i == avd.indexTrigger1)
                            newColor = AVD.trigger1Color;
                        if (i == avd.indexTrigger2)
                            newColor = AVD.trigger2Color;
                    }
                    for (c = 0; c < colors.Length; c++)
                        colors[c] = newColor;
                    meshFilters[i].mesh.colors = colors; // slowest line in loop
                }

                // BENCHMARK END
                stopWatch.Stop();
                avgUpdateTimes[avgUpdateTimes.Length - 1] = stopWatch.Elapsed.TotalMilliseconds;
                Array.Copy(avgUpdateTimes, 1, avgUpdateTimes, 0, avgUpdateTimes.Length - 1); // shift array
                avgUpdateTime = 0;
                for (int i = 0; i < avgUpdateTimes.Length; i++)
                    avgUpdateTime += avgUpdateTimes[i];
                avgUpdateTime = avgUpdateTime / avgUpdateTimes.Length;

                jsBenchmark.val = $"Update Benchmark: {avgUpdateTime:F4} ms";
            }
            catch (Exception e) { SuperController.LogError("Exception: " + e); }
        }

        void OnDestroy()
        {
            //Debug.Log($"{name} OnDestroy()");
            //Debug.Log($"{name} OnDestroy() meshRenderers.Lengt: {meshRenderers.Length}  meshFilters.Lengt: {meshFilters.Length}   cubes.Lengt: {cubes.Length}  cubeMaterial: {cubeMaterial}");
            int i;
            // meshRenderers
            for (i = 0; i < meshRenderers.Length; i++)
            {
                Destroy(meshRenderers[i]);
                meshRenderers[i] = null;
            }
            meshRenderers = null;

            // meshFilters
            for (i = 0; i < meshFilters.Length; i++)
            {
                //Debug.Log($"{name} OnDestroy() meshFilters[i].mesh: {meshFilters[i].mesh}");
                if (meshFilters[i].mesh != null)
                {
                    meshFilters[i].mesh.Clear();
                    Destroy(meshFilters[i].mesh);
                    meshFilters[i].mesh = null;
                }
                Destroy(meshFilters[i]);
                meshFilters[i] = null;
            }
            meshFilters = null;

            // cubes[] Gameobjects
            for (i = 0; i<cubes.Length; i++)
            {
                Destroy(cubes[i]);
                cubes[i] = null;
            }
            cubes = null;

            // CubeMaterial
            if (cubeMaterial != null)
            {
                Destroy(cubeMaterial);
                cubeMaterial = null;
            }

            avd = null;
            avgUpdateTimes = null;
        }

        //////////////////////////////////////////////////////////////////////////////////
        // AVD Synchronization - called by SendMessage from Master
        //////////////////////////////////////////////////////////////////////////////////
        public void Sync(float[] syncdata)
        {
            try
            {
                avd.UnpackSyncdata(syncdata);
            }
            catch (Exception e) { SuperController.LogError("Exception: " + e); }
        }

        //////////////////////////////////////////////////////////////////////////////////
        // UI Callbacks
        //////////////////////////////////////////////////////////////////////////////////
        private void RebuildSpectrum(float s) { CreateVisualization(); }
        private void SetHiColor(float h, float s, float v) { colorHiRGB = Color.HSVToRGB(h, s, v); }
        private void SetLoColor(float h, float s, float v) { colorLoRGB = Color.HSVToRGB(h, s, v); }
        private void SetLimitSamples(float v)
        {
            for (int i = 0; i < cubes.Length; i++)
            {
                if (i < (int)v)
                    cubes[i].SetActive(true);
                else
                    cubes[i].SetActive(false);
            }
        }
    }
}