﻿// ***********************************************************************************
// SallySoundFX
// ***********************************************************************************
// Created by Sally Whitemane on VAM Hub
// https://www.patreon.com/sallyw
// Licensed under CC BY-SA - https://creativecommons.org/licenses/by-sa/4.0/

using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using MVR.FileManagementSecure;

namespace Sally
{
    public class AVSLAVE_Tiles : MVRScript
    {
        // UI - LEFT - GEOMETRY
        private JSONStorableStringChooser jStringChooserGeometry;
        private Dictionary<string, Action> geometryChoicesDict;
        private Action geometryLogicAction;
        private JSONStorableFloat jfTileScaleMin;
        private UIDynamicSlider tileScaleSliderMin;
        private JSONStorableFloat jfTileScaleMax;
        private UIDynamicSlider tileScaleSliderMax;
        // UI - LEFT - COLOR ALPHA
        private JSONStorableFloat jfAlpha;
        // UI - LEFT - COLOR
        private JSONStorableStringChooser jStringChooserColor;
        private Dictionary<string, Action> colorChoicesDict;
        private Action colorLogicAction;
        private JSONStorableColor jColorHi;
        private JSONStorableColor jColorLo;

        // UI - RIGHT
        private JSONStorableString jsBenchmark;
        private JSONStorableStringChooser jShader;
        private List<string> shaderChoices;
        private JSONStorableUrl jTextureURL;
        private JSONStorableFloat jTextureRotation;
        private JSONStorableFloat jfXSize;
        private JSONStorableFloat jfZSize;
        private JSONStorableFloat jfScale;
        private JSONStorableFloat jfRenderQueue;

        // BENCHMARK keep namespace, can cause VAM compilation error
        private System.Diagnostics.Stopwatch stopWatch = new System.Diagnostics.Stopwatch();
        private double[] avgUpdateTimes = new double[80];
        private double avgUpdateTime = 0;

        // AUDIO VISUALIZATION DATA
        private AVD avd = new AVD();

        // RESOURCES
        private string texturePath;

        // VISUALIZATION
        private int xSize = 8; // x-width in number of tiles
        private int zSize = 8; // z-depth in number of tiles

        // VISUALIZATION TILES
        private GameObject goTiles;
        private Mesh mesh;
        private Shader shader;
        private Texture2D texture;
        private Material material;
        private Vector3[] vertices;
        private Color[] colors;
        // vertex colors
        private Color colorWhite = Color.white;   // default color
        private Color colorLoRGB = Color.cyan;    // stores RGB-version of jColorLo
        private Color colorHiRGB = Color.magenta; // stores RGB-version of jColorHi
        private Color loopColor  = Color.white;   // optimization

        private float halfTileScaleMin = 0.25f;
        private float halfTileScaleMax = 0.5f;
        private float globalHue  = 0.0f;
        private float deltaTimeCounter = 0f;      // used to limit Visualization framerate
        private float counter = 0;
        //private float[] sin1x = LUT.CreateSinWave(AVD.SAMPLES, 1, true);
        public override void Init()
        {
            try
            {
                //Debug.Log($"{name} Init()");

                AVSlave.GetPluginPath(this);

                if (AVSlave.GetMasterScript(this) == null)
                    SuperController.LogMessage($"{name} could not find SallySoundFX Plugin to synchronize. It will retry automatically.");

                texturePath = AVSlave.GetPluginPath(this) + "/textures";
                //Debug.Log($"{name} texturePath: {texturePath}");

                //////////////////////////////////////////////////////////////////////////////////
                // UI LEFT
                //////////////////////////////////////////////////////////////////////////////////
                UIDynamicSlider slider;
                UIDynamicButton button;

                // Button/Label Geometry Logic
                button = CreateButton("Geometry Logic:");
                button.button.interactable = false;
                button.buttonText.fontSize = 32;
                button.buttonText.fontStyle = FontStyle.Bold;
                button.enabled = false;
                // StringChoose Geometry Logic
                geometryChoicesDict = new Dictionary<string, Action>();
                geometryChoicesDict.Add("None (fastest)", GeometryLogicNone);
                geometryChoicesDict.Add("Scale Tiles by Bands", GeometryLogicScaleTilesByBands);
                geometryChoicesDict.Add("Scale Tiles by Samples", GeometryLogicScaleTilesBySamples);
                // DEFAULT: None (fastest)
                jStringChooserGeometry = new JSONStorableStringChooser("StringChooserGeometryLogic", geometryChoicesDict.Keys.ToList(), "None (fastest)", "", SetGeometryLogic);
                RegisterStringChooser(jStringChooserGeometry);
                CreateScrollablePopup(jStringChooserGeometry).labelWidth = 0;
                // Set Geometry Logic at end of Init() after Visualization was created

                // Slider Tile Scale Min
                jfTileScaleMin = new JSONStorableFloat("Tile Scale Minimum", 0.5f, 0.1f, 4f);
                RegisterFloat(jfTileScaleMin);
                tileScaleSliderMin = CreateSlider(jfTileScaleMin);
                jfTileScaleMin.setCallbackFunction += SetTileScaleMin;
                SetTileScaleMin(jfTileScaleMin.val);
                // Slider Tile Scale Max
                jfTileScaleMax = new JSONStorableFloat("Tile Scale Maximum", 1.0f, 0.1f, 4f);
                RegisterFloat(jfTileScaleMax);
                tileScaleSliderMax = CreateSlider(jfTileScaleMax);
                jfTileScaleMax.setCallbackFunction += SetTileScaleMax;
                SetTileScaleMax(jfTileScaleMax.val);

                // Button/Label Color Logic
                button = CreateButton("Color Logic:");
                button.button.interactable = false;
                button.buttonText.fontSize = 32;
                button.buttonText.fontStyle = FontStyle.Bold;
                button.enabled = false;
                // StringChooser Color logic
                colorChoicesDict = new Dictionary<string, Action>();
                colorChoicesDict.Add("None (White)", ColorLogicNoneWhite);
                colorChoicesDict.Add("Freq-Bands [Hi/Lo Color]", ColorLogic1);
                colorChoicesDict.Add("Freq-Samples [Hi/Lo Color]", ColorLogic2);
                colorChoicesDict.Add("Freq-Bands Push [Hi/Lo Color]", ColorLogic3);
                colorChoicesDict.Add("Rainbow Bands", ColorLogic4);
                colorChoicesDict.Add("Rainbow Tiles Samples Peak", ColorLogic5);
                colorChoicesDict.Add("SSFX Trigger 1 [Hi/Lo Color]", ColorLogic6);
                colorChoicesDict.Add("SSFX Trigger 2 [Hi/Lo Color]", ColorLogic7);
                colorChoicesDict.Add("SSFX Trigger 1 Push [Hi/Lo Color]", ColorLogic8);
                // DEFAULT: Freq-Bands [Hi/Lo Color]
                jStringChooserColor = new JSONStorableStringChooser("StringChooserColorLogic", colorChoicesDict.Keys.ToList(), "Freq-Bands [Hi/Lo Color]", "", SetColorLogic);
                RegisterStringChooser(jStringChooserColor);
                CreateScrollablePopup(jStringChooserColor).labelWidth = 0;
                // Set Color Logic at end of Init() after Visualization was created
                
                // Slider Alpha
                jfAlpha = new JSONStorableFloat("Transparency (Alpha)", 1.0f, SetVertexColorAlpha, 0.0f, 1.0f);
                RegisterFloat(jfAlpha);
                CreateSlider(jfAlpha);

                // Color Picker High
                jColorHi = new JSONStorableColor("Color High", HSVColorPicker.RGBToHSV(1f, 0f, 1f), SetColorHi);
                RegisterColor(jColorHi);
                CreateColorPicker(jColorHi);
                SetColorHi(jColorHi.val.H, jColorHi.val.S, jColorHi.val.V);
                UI.FixInitialColorPickerSelectorPosition(jColorHi);

                // Color Picker Low
                jColorLo = new JSONStorableColor("Color Low", HSVColorPicker.RGBToHSV(0.0625f, 0.0625f, 0.0625f), SetColorLo);
                RegisterColor(jColorLo);
                CreateColorPicker(jColorLo);
                SetColorLo(jColorLo.val.H, jColorLo.val.S, jColorLo.val.V);
                UI.FixInitialColorPickerSelectorPosition(jColorLo);

                //////////////////////////////////////////////////////////////////////////////////
                // UI RIGHT
                //////////////////////////////////////////////////////////////////////////////////

                // Benchmark
                jsBenchmark = new JSONStorableString("Benchmark", "no data");
                UIDynamicTextField tf = CreateTextField(jsBenchmark, true);
                LayoutElement le = tf.GetComponent<LayoutElement>();
                le.minHeight = 40.0f;
                le.preferredHeight = 40.0f;

                // Button/Label Shaders
                button = CreateButton("Shaders: <size=24><color=#A00000>(experimental)</color></size>", true);
                button.button.interactable = false;
                button.buttonText.fontSize = 32;
                button.buttonText.fontStyle = FontStyle.Bold;
                button.enabled = false;

                // StringChooser Shader
                // https://answers.unity.com/questions/1353091/is-there-any-way-to-create-and-load-a-shader-sourc.html
                // NOT working:
                // Unlit/Texture
                // Legacy Shaders/VertexLit
                shaderChoices = new List<string>();
                shaderChoices.Add("UI/Default");                      // renderQueue: 3000 [vertex color][unlit][no culling]
                shaderChoices.Add("Sprites/Default");                 // renderQueue: 3000 [vertex color][unlit][no culling]
                shaderChoices.Add("Particles/Alpha Blended");         // renderQueue: 3000 [vertex color][unlit][no culling] => darker from blending on black background
                shaderChoices.Add("Custom/Subsurface/EmissiveGlow");  // renderQueue: ???? [no vertex color][culled]
                //shaderChoices.Add("Battlehub/RTHandles/Grid");        // renderQueue: ???? [unlit][no culling] no texture?
                //shaderChoices.Add("Battlehub/RTHandles/VertexColor"); // renderQueue: ???? [unlit][no culling] no texture?
                shaderChoices.Add("Standard");                        // renderQueue: 2000 [mat color][lit][culled]
                jShader = new JSONStorableStringChooser("stringChooserShader", shaderChoices, "UI/Default", "", SetShader);
                RegisterStringChooser(jShader);
                //CreatePopup(jShader, true).labelWidth = 0;
                CreateScrollablePopup(jShader, true).labelWidth = 0;

                // Button Load Texture
                jTextureURL = new JSONStorableUrl("TextureURL", "");
                RegisterUrl(jTextureURL);
                if (!FileManagerSecure.DirectoryExists(texturePath))
                    FileManagerSecure.CreateDirectory(texturePath);
                button = CreateButton("Load Texture", true);
                button.buttonColor = UI.redWarningBackground;
                button.button.onClick.AddListener(() =>
                {
                    List<ShortCut> shortcuts = FileManagerSecure.GetShortCutsForDirectory("");
                    ShortCut pluginDefault = new ShortCut();
                    pluginDefault.path = texturePath;
                    pluginDefault.displayName = texturePath;
                    shortcuts.Insert(0, pluginDefault);

                    SuperController.singleton.GetMediaPathDialog(LoadTexture, "", texturePath);
                });

                // Botton Rotate Texture
                jTextureRotation = new JSONStorableFloat("Texture Rotation", 0f, 0f, 360f);
                RegisterFloat(jTextureRotation);
                button = CreateButton("Rotate Texture", true);
                button.button.onClick.AddListener(() => { RotateTexture(); });

                // Slider X Size
                jfXSize = jfXSize = new JSONStorableFloat("X Size", 8f, 1, 16f);
                RegisterFloat(jfXSize);
                slider = CreateSlider(jfXSize, true);
                slider.Configure(jfXSize.name, jfXSize.min, jfXSize.max, jfXSize.val, true, "F0");
                slider.GetComponentInChildren<Image>().color = UI.redWarningBackground;
                //slider.quickButtonM1.gameObject.SetActive(false); // -1
                //slider.quickButtonM2.gameObject.SetActive(false); // -10
                slider.quickButtonM3.gameObject.SetActive(false); // -100
                slider.quickButtonM4.gameObject.SetActive(false); // -1000
                //slider.quickButtonP1.gameObject.SetActive(false); // +1
                //slider.quickButtonP2.gameObject.SetActive(false); // +10
                slider.quickButtonP3.gameObject.SetActive(false); // +100
                slider.quickButtonP4.gameObject.SetActive(false); // +1000
                jfXSize.setCallbackFunction += SetXSize;

                // Slider Z Size
                jfZSize = new JSONStorableFloat("Z Size", 8f, 1, 16f);
                RegisterFloat(jfZSize);
                slider = CreateSlider(jfZSize, true);
                slider.Configure(jfZSize.name, jfZSize.min, jfZSize.max, jfZSize.val, true, "F0");
                slider.GetComponentInChildren<Image>().color = UI.redWarningBackground;
                //slider.quickButtonM1.gameObject.SetActive(false); // -1
                //slider.quickButtonM2.gameObject.SetActive(false); // -10
                slider.quickButtonM3.gameObject.SetActive(false); // -100
                slider.quickButtonM4.gameObject.SetActive(false); // -1000
                //slider.quickButtonP1.gameObject.SetActive(false); // +1
                //slider.quickButtonP2.gameObject.SetActive(false); // +10
                slider.quickButtonP3.gameObject.SetActive(false); // +100
                slider.quickButtonP4.gameObject.SetActive(false); // +1000
                jfZSize.setCallbackFunction += SetZSize;

                // Slider Scale
                jfScale = new JSONStorableFloat("Scale", 0.5f, 0.02f, 2.0f);
                RegisterFloat(jfScale);
                slider = CreateSlider(jfScale, true);
                slider.Configure(jfScale.name, jfScale.min, jfScale.max, jfScale.val, true, "F4");
                jfScale.setCallbackFunction += SetScale;

                // Render Queue
                jfRenderQueue = new JSONStorableFloat("Render Queue", 2400f, -1, 5000f);
                RegisterFloat(jfRenderQueue);
                CreateSlider(jfRenderQueue, true);
                jfRenderQueue.setCallbackFunction += SetRenderQueue;

                //////////////////////////////////////////////////////////////////////////////////
                // VISUALIZATION
                //////////////////////////////////////////////////////////////////////////////////
                CreateVisualization();

                SetColorLogic(jStringChooserColor.val);
                SetGeometryLogic(jStringChooserGeometry.val);
            }
            catch (Exception e) { SuperController.LogError("Exception: " + e); }
        }

        private void LoadTexture(string path)
        {
            //Debug.Log($"{name} LoadTexture() path: {path}");
            // example path: Author.VarPackageName.Version:/Custom/Scripts/SallySoundFX/textures/light_1_greyscale_512x512.png
            if (path == "")  return; // nothing was selected
            ImageLoaderThreaded.QueuedImage queuedImage = new ImageLoaderThreaded.QueuedImage();
            queuedImage.imgPath = path;
            queuedImage.forceReload = false;
            queuedImage.createMipMaps = true;
            queuedImage.isNormalMap = false;
            queuedImage.isThumbnail = false;
            queuedImage.linear = false; // linear or sRGB color space https://docs.unity3d.com/Manual/LinearRendering-LinearOrGammaWorkflow.html
            queuedImage.createAlphaFromGrayscale = false;
            queuedImage.compress = false;
            queuedImage.callback = new ImageLoaderThreaded.ImageLoaderCallback(OnImageLoaded);
            ImageLoaderThreaded.singleton.QueueImage(queuedImage);
        }

        // ImageLoaderThreaded Callback
        private void OnImageLoaded(ImageLoaderThreaded.QueuedImage queuedImage)
        {
            if (queuedImage != null)
            {
                // Destroy old texture <coded removed - safe? Memory leak if not destroyed?>
                // Only destroy if not in use on another AVSlave!
                //TextureRaw.DestroyIfUnused(this, texture); // < breaks texture loading for unknown reason

                //Debug.Log($"{name} OnImageLoaded()  queuedImage.tex.name: {queuedImage.tex.name}");
                material.SetTexture("_MainTex", queuedImage.tex);
                texture = queuedImage.tex;
                //Debug.Log($"{name} OnImageLoaded()  texture.name: {texture.name}");

                // Update texture Url to be saved with scene
                jTextureURL.val = queuedImage.imgPath;
                //Debug.Log($"{name} OnImageLoaded() jTextureURL.val {jTextureURL.val}");
            }
        }

        // Start is called once before Update or FixedUpdate is called and after Init()
        void Start()
        {
            try
            {
                // Hide parts of UI
                if (geometryLogicAction == GeometryLogicNone)
                {
                    tileScaleSliderMin.gameObject.SetActive(false);
                    tileScaleSliderMax.gameObject.SetActive(false);
                }

                //Debug.Log($"{name} Start() jTextureURL.val: {jTextureURL.val}");
                if (jTextureURL.val != "")
                    LoadTexture(jTextureURL.val);
            } 
            catch (Exception e) { SuperController.LogError("Exception: " + e); }
        }

        private void CreateVisualization()
        {
            //Debug.Log($"{name} CreateVisualization()");

            // TEXTURE
            // Load raw DXT1 texture from C# byte array, fallback in case AssetBundle fails to load
            texture = new Texture2D(128, 128, TextureFormat.DXT1, false);
            if (texture == null)
            {
                SuperController.LogError($"{name} Could not create texture. (texture == null)");
                return;
            }
            texture.LoadRawTextureData(TextureRaw.tileDXT1);
            texture.Apply();
            texture.name = "ssfx_default";

            // SHADER
            shader = Shader.Find(jShader.val); // UI/Default  V2.00 Sprites/Default
            if (shader == null)
            {
                SuperController.LogError($"{name} Could not find shader. (shader == null)");
                return;
            }

            // GAMEOBJECT
            goTiles = new GameObject("goTiles", typeof(MeshFilter), typeof(MeshRenderer));

            // CREATE MESH
            goTiles.GetComponent<MeshFilter>().mesh = new Mesh();
            mesh = goTiles.GetComponent<MeshFilter>().mesh;

            CreateMeshWithSize();

            // MATERIAL
            material = new Material(shader);
            if (material == null)
            {
                SuperController.LogError($"{name} Could not create Material. (material == null)");
                return;
            }
            // Make sure custom shaders diffuse texture property is named "MainTex" too
            material.SetTexture("_MainTex", texture);
            //material.renderQueue = 1;
            goTiles.GetComponent<MeshRenderer>().material = material;


            // POSITION & TRANSLATION
            goTiles.transform.SetParent(containingAtom.mainController.transform);
            goTiles.transform.localScale = new Vector3(jfScale.val, jfScale.val, jfScale.val);
            SyncToAtom();
        }

        private void CreateMeshWithSize()
        {
            mesh.Clear();
            vertices = new Vector3[xSize * zSize * 4];
            Vector2[] uv = new Vector2[xSize * zSize * 4];
            int[] triangles = new int[xSize * zSize * 6];
            colors = new Color[vertices.Length];

            //Debug.Log($"{name} vertices.Length: " + vertices.Length);
            //Debug.Log($"{name} uv.Length: " + uv.Length);
            //Debug.Log($"{name} triangles.Length: " + triangles.Length);
            //Debug.Log($"{name} colors.Length: " + colors.Length);

            int vCount = 0;
            int vIndex = 0;
            int x;
            for (int z = 0; z < zSize; z++)
            {
                for (x = 0; x < xSize; x++)
                {
                    vertices[vCount] = new Vector3(x, 0f, z + 1f);
                    vertices[vCount + 1] = new Vector3(x + 1f, 0f, z + 1f);
                    vertices[vCount + 2] = new Vector3(x, 0f, z + 0f);
                    vertices[vCount + 3] = new Vector3(x + 1f, 0f, z + 0f);

                    if (jTextureRotation.val == 0f)
                    {
                        uv[vCount] = new Vector2(0f, 1f);
                        uv[vCount + 1] = new Vector2(1f, 1f);
                        uv[vCount + 2] = new Vector2(0f, 0f);
                        uv[vCount + 3] = new Vector2(1f, 0f);
                    }
                    else if (jTextureRotation.val == 90f)
                    {
                        uv[vCount] = new Vector2(0f, 0f);
                        uv[vCount + 1] = new Vector2(0f, 1f);
                        uv[vCount + 2] = new Vector2(1f, 0f);
                        uv[vCount + 3] = new Vector2(1f, 1f);
                    }
                    else if (jTextureRotation.val == 180f)
                    {
                        uv[vCount] = new Vector2(1f, 0f);
                        uv[vCount + 1] = new Vector2(0f, 0f);
                        uv[vCount + 2] = new Vector2(1f, 1f);
                        uv[vCount + 3] = new Vector2(0f, 1f);
                    }
                    else
                    {
                        uv[vCount] = new Vector2(1f, 1f);
                        uv[vCount + 1] = new Vector2(1f, 0f);
                        uv[vCount + 2] = new Vector2(0f, 1f);
                        uv[vCount + 3] = new Vector2(0f, 0f);
                    }

                    colors[vCount] = colorWhite;
                    colors[vCount + 1] = colorWhite;
                    colors[vCount + 2] = colorWhite;
                    colors[vCount + 3] = colorWhite;

                    triangles[vIndex] = vCount;
                    triangles[vIndex + 1] = vCount + 1;
                    triangles[vIndex + 2] = vCount + 2;
                    triangles[vIndex + 3] = vCount + 2;
                    triangles[vIndex + 4] = vCount + 1;
                    triangles[vIndex + 5] = vCount + 3;
                    vCount += 4;
                    vIndex += 6;
                }
            }

            mesh.vertices = vertices;
            mesh.uv = uv;
            mesh.triangles = triangles;
            mesh.colors = colors;
        }

        // Update is called with each rendered frame by Unity
        private void Update()
        {
            try
            {
                // BENCHMARK START
                stopWatch.Reset();
                stopWatch.Start();

                if (!AVSlave.IsSynchronized(this))
                    return;

                // Color logic
                //if (colorLogicAction != null)
                colorLogicAction();

                // Geometry logic
                geometryLogicAction();

                // BENCHMARK END
                stopWatch.Stop();
                avgUpdateTimes[avgUpdateTimes.Length - 1] = stopWatch.Elapsed.TotalMilliseconds;
                Array.Copy(avgUpdateTimes, 1, avgUpdateTimes, 0, avgUpdateTimes.Length - 1); // shift array
                avgUpdateTime = 0;
                for (int i = 0; i < avgUpdateTimes.Length; i++)
                    avgUpdateTime += avgUpdateTimes[i];
                avgUpdateTime = avgUpdateTime / avgUpdateTimes.Length;

                jsBenchmark.val = $"Update Benchmark: {avgUpdateTime:F4} ms";
            }
            catch (Exception e) { SuperController.LogError("Exception: " + e); }
        }

        // OnDestroy is where you should put any cleanup
        // if you registered objects to supercontroller or atom, you should unregister them here
        void OnDestroy()
        {
            //Debug.Log($"{name} OnDestroy()");
            avgUpdateTimes = null;
            avd = null;
            texturePath = null;

            if (goTiles != null)
            {
                Destroy(goTiles);
                goTiles = null;
            }

            if (mesh != null)
            {
                mesh.Clear();
                Destroy(mesh);
                mesh = null;
            }

            if (shader != null)
            {
                // destroys VAM default shader
                //Destroy(shader);
                shader = null;
            }

            if (texture != null)
            {
                TextureRaw.DestroyIfUnused(this, texture);
                //Destroy(texture);
                texture = null;
            }

            if (material != null)
            {
                Destroy(material);
                material = null;
            }

            vertices = null;
            colors = null;
        }

        //////////////////////////////////////////////////////////////////////////////////
        // UI Callbacks
        //////////////////////////////////////////////////////////////////////////////////
        
        private void SetVertexColorAlpha(float v)
        {
            colorWhite.a = v;
            colorLoRGB.a = v;
            colorHiRGB.a = v;
            loopColor.a = v;
            //Mesh mesh = goTiles.GetComponent<MeshFilter>().mesh;
            for (int i = 0; i < colors.Length; i++)
            {
                colors[i].a = v;
            }
            mesh.colors = colors;
        }
        
        private void SetShader(string s)
        {
            Shader newShader = Shader.Find(s);
            if (newShader == null)
            {
                SuperController.LogError($"{name} SetShader() could not find shader. (shader == null)");
                return;
            }

            if (material == null)
            {
                // Destroy current material
                Destroy(material);
                material = null;
            }
            // Create new Material from new Shader
            material = new Material(newShader);
            material.mainTexture = texture;
            
            // Apply to GameObject
            goTiles.GetComponent<MeshRenderer>().material = material;
        }

        private void SetRenderQueue(float v)
        {
            jfRenderQueue.val = (int)v;
            material.renderQueue = (int)v;
        }

        private void SetXSize(float v)
        {
            xSize = (int)v;
            jfXSize.val = xSize;
            CreateMeshWithSize();
            if (goTiles != null)
                SyncToAtom();
        }

        private void SetZSize(float v)
        {
            zSize = (int)v;
            jfZSize.val = zSize;
            CreateMeshWithSize();
            if (goTiles != null)
                SyncToAtom();
        }

        private void SetScale(float v)
        {
            if (goTiles != null)
            {
                goTiles.transform.localScale = new Vector3(v, v, v);
                SyncToAtom();
            }
        }

        private void SetTileScaleMin(float v)
        {
            halfTileScaleMin = v * 0.5f;
            if (jfTileScaleMax != null && jfTileScaleMax.val < v)
            {
                // Maximum is < Minimum, fix that
                jfTileScaleMax.val = v;
            }
        }

        private void SetTileScaleMax(float v)
        {
            halfTileScaleMax = v * 0.5f;
            if (jfTileScaleMin != null && jfTileScaleMin.val > v)
            {
                // Minimum is > Maximum, fix that
                jfTileScaleMin.val = v;
            }
        }

        private void SetColorLogic(string s)
        {
            //Debug.Log($"{name}  SetColorLogic()  s: {s}");
            colorLogicAction = colorChoicesDict[s];
            if (colorLogicAction == ColorLogicNoneWhite)
                SetMeshColorWhite();
        }

        private void SetGeometryLogic(string s)
        {
            //Debug.Log($"{name}  SetGeometryLogic()  s: {s}");
            geometryLogicAction = geometryChoicesDict[s];
            if (geometryLogicAction == GeometryLogicNone)
            {
                //Debug.Log($"{name} SetGeometryLogic() Mesh reset");
                ResetMeshGeometry();
                tileScaleSliderMin.gameObject.SetActive(false);
                tileScaleSliderMax.gameObject.SetActive(false);                
            }
            else if (geometryLogicAction == GeometryLogicScaleTilesByBands)
            {
                tileScaleSliderMin.gameObject.SetActive(true);
                tileScaleSliderMax.gameObject.SetActive(true);
            }
            else if (geometryLogicAction == GeometryLogicScaleTilesBySamples)
            {
                tileScaleSliderMin.gameObject.SetActive(true);
                tileScaleSliderMax.gameObject.SetActive(true);
            }
        }

        private void SetColorHi(float h, float s, float v) { colorHiRGB = Color.HSVToRGB(h, s, v); }
        private void SetColorLo(float h, float s, float v) { colorLoRGB = Color.HSVToRGB(h, s, v); }

        //////////////////////////////////////////////////////////////////////////////////
        // Helper functions
        //////////////////////////////////////////////////////////////////////////////////

        private void SyncToAtom()
        {
            // move goTiles to atom
            goTiles.transform.position = containingAtom.mainController.transform.position;
            goTiles.transform.rotation = containingAtom.mainController.transform.rotation;
            // center goTiles on atom
            goTiles.transform.Translate(new Vector3(-xSize * 0.5f * jfScale.val, 0.0f, -zSize * 0.5f * jfScale.val));
        }

        private void SetMeshColorWhite()
        {
            if (colors == null)
                return;
            for (int i = 0; i < colors.Length; i++)
                colors[i] = colorWhite;
            goTiles.GetComponent<MeshFilter>().mesh.colors = colors;
        }

        private void RotateTexture()
        {
            jTextureRotation.val += 90f;
            if (jTextureRotation.val > 271f)
                jTextureRotation.val = 0f;

            Mesh mesh = goTiles.GetComponent<MeshFilter>().mesh;
            Vector2[] uv = new Vector2[xSize * zSize * 4];
            int vCount = 0;
            for (int z = 0; z < zSize; z++)
                for (int x = 0; x < xSize; x++)
                {
                    if (jTextureRotation.val == 0f)
                    {
                        uv[vCount] = new Vector2(0f, 1f);
                        uv[vCount + 1] = new Vector2(1f, 1f);
                        uv[vCount + 2] = new Vector2(0f, 0f);
                        uv[vCount + 3] = new Vector2(1f, 0f);
                    }
                    else if (jTextureRotation.val == 90f)
                    {
                        uv[vCount] = new Vector2(0f, 0f);
                        uv[vCount + 1] = new Vector2(0f, 1f);
                        uv[vCount + 2] = new Vector2(1f, 0f);
                        uv[vCount + 3] = new Vector2(1f, 1f);
                    }
                    else if (jTextureRotation.val == 180f)
                    {
                        uv[vCount] = new Vector2(1f, 0f);
                        uv[vCount + 1] = new Vector2(0f, 0f);
                        uv[vCount + 2] = new Vector2(1f, 1f);
                        uv[vCount + 3] = new Vector2(0f, 1f);
                    }
                    else
                    {
                        uv[vCount] = new Vector2(1f, 1f);
                        uv[vCount + 1] = new Vector2(1f, 0f);
                        uv[vCount + 2] = new Vector2(0f, 1f);
                        uv[vCount + 3] = new Vector2(0f, 0f);
                    }
                    vCount += 4;
                }
            mesh.uv = uv;
        }

        //////////////////////////////////////////////////////////////////////////////////
        // Color Logic
        //////////////////////////////////////////////////////////////////////////////////

        private void ColorLogicNoneWhite() { }// only colored white once by StringChooser Callback -> fast

        private void ColorLogic1() // "Freq-Bands [Hi/Lo Color]"
        {
            // Low / Bands
            int band = 0;
            float v;
            float v_inverted;
            for (int i = 0; i < colors.Length; i += 4)
            {
                v = avd.bands[band];
                v_inverted = 1f - v;
                colors[i].r = colorLoRGB.r * v_inverted + colorHiRGB.r * v;
                colors[i].g = colorLoRGB.g * v_inverted + colorHiRGB.g * v;
                colors[i].b = colorLoRGB.b * v_inverted + colorHiRGB.b * v;
                colors[i + 1] = colors[i + 2] = colors[i + 3] = colors[i];
                band++;
                if (band >= avd.bands.Length)
                    band = 0;
            }
            goTiles.GetComponent<MeshFilter>().mesh.colors = colors;
        }

        private void ColorLogic2() // "Freq-Samples [Hi/Lo Color]"
        {
            // Samples
            int sample = 0;
            float v;
            float v_inverted;
            for (int i = 0; i < colors.Length; i += 4)
            {
                v = avd.samples[sample];
                v_inverted = 1f - v;
                colors[i].r = colorLoRGB.r * v_inverted + colorHiRGB.r * v;
                colors[i].g = colorLoRGB.g * v_inverted + colorHiRGB.g * v;
                colors[i].b = colorLoRGB.b * v_inverted + colorHiRGB.b * v;
                colors[i + 1] = colors[i + 2] = colors[i + 3] = colors[i];
                sample++;
                if (sample >= avd.samples.Length)
                    sample = 0;
            }
            goTiles.GetComponent<MeshFilter>().mesh.colors = colors;
        }

        private void ColorLogic3() // "Freq-Bands Push [Hi/Lo Color]"
        {
            int band = 0;
            int x = 0;
            int z = 0;
            int vCount = 0;
            float value = avd.bands[band];

            for (z = 0; z < zSize; z++)
            {
                for (x = 0; x < xSize; x++)
                {
                    if (x < value * xSize)
                    {
                        colors[vCount] = loopColor;
                        colors[vCount + 1] = loopColor;
                        colors[vCount + 2] = loopColor;
                        colors[vCount + 3] = loopColor;
                    }
                    vCount += 4;
                }
                band++;
                if (band >= avd.bands.Length)
                    band = 0;
                value = avd.bands[band];
                loopColor.r = colorHiRGB.r;
                loopColor.g = colorHiRGB.g;
                loopColor.b = colorHiRGB.b;
            }

            // framerate limit (to fade color at fixed rate)
            deltaTimeCounter += Time.deltaTime;
            if (deltaTimeCounter > 1f / 30f)
            {
                deltaTimeCounter = deltaTimeCounter - 1f / 30f;

                for (int i = 0; i < colors.Length; i += 4)
                {
                    // // fade away color over time (* 0.95f), keep colorLoRGB minimum
                    if (colors[i].r <= colorLoRGB.r)
                        colors[i].r = colorLoRGB.r;
                    else
                        colors[i].r = colors[i].r * 0.9f;

                    if (colors[i].g <= colorLoRGB.g)
                        colors[i].g = colorLoRGB.g;
                    else
                        colors[i].g = colors[i].g * 0.9f;

                    if (colors[i].b <= colorLoRGB.b)
                        colors[i].b = colorLoRGB.b;
                    else
                        colors[i].b = colors[i].b * 0.9f;
                    colors[i + 1] = colors[i + 2] = colors[i + 3] = colors[i];
                }
            }
            goTiles.GetComponent<MeshFilter>().mesh.colors = colors;
        }

        private void ColorLogic4() // "Rainbow Bands"
        {
            int band = 0;
            float hue = 0f;
            float hue_offset = 1f / xSize;
            float v;
            float alpha = jfAlpha.val;
            for (int i = 0; i < colors.Length; i += 4)
            {
                //v = avd.bands[band];
                v = 0.02f + avd.bands[band] * 0.98f;
                colors[i] = Color.HSVToRGB(hue, 1f, v);
                colors[i].a = alpha;
                colors[i + 1] = colors[i + 2] = colors[i + 3] = colors[i];
                hue += hue_offset;
                if (hue >= 1f)
                    hue = 0f;
                band++;
                if (band >= avd.bands.Length)
                    band = 0;
            }
            goTiles.GetComponent<MeshFilter>().mesh.colors = colors;
        }

        private void ColorLogic5() // "Rainbow Tiles Samples Peak"
        {
            int sample = 0;
            float localHue = 0; // change per tile
            float hue = 0;
            float v;
            float alpha = jfAlpha.val;
            for (int i = 0; i < colors.Length; i += 4)
            {
                v = avd.samples_peak[sample];
                hue = globalHue + localHue;
                if (hue > 1f)
                    hue = hue - 1f;

                colors[i] = Color.HSVToRGB(hue, 1f, v);
                colors[i].a = alpha;
                colors[i + 1] = colors[i + 2] = colors[i + 3] = colors[i];

                localHue += 0.10f;
                if (localHue > 1f)
                    localHue = localHue - 1f;

                sample++;
                if (sample >= avd.samples_peak.Length)
                    sample = 0;
            }

            // framerate limit
            deltaTimeCounter += Time.deltaTime;
            if (deltaTimeCounter > 1f / 30f)
            {
                deltaTimeCounter = deltaTimeCounter - 1f / 30f;
                globalHue += 0.01f;
                if (globalHue > 1f)
                    globalHue = globalHue - 1f;
            }
            goTiles.GetComponent<MeshFilter>().mesh.colors = colors;
        }
                
        private void ColorLogic6() // "SSFX Trigger 1 [Hi/Lo Color]"
        {
            // set all to sample value of SSFX trigger index 1/2
            float v = avd.samples[avd.indexTrigger1];
            float v_inverted = 1f - v;
            loopColor.r = colorLoRGB.r * v_inverted + colorHiRGB.r * v;
            loopColor.g = colorLoRGB.g * v_inverted + colorHiRGB.g * v;
            loopColor.b = colorLoRGB.b * v_inverted + colorHiRGB.b * v;
            for (int i = 0; i < colors.Length; i++)
                colors[i] = loopColor;
            goTiles.GetComponent<MeshFilter>().mesh.colors = colors;
        }

        private void ColorLogic7() // "SSFX Trigger 2 [Hi/Lo Color]"
        {
            // set all to sample value of SSFX trigger index 1/2
            float v = avd.samples[avd.indexTrigger2];
            float v_inverted = 1f - v;
            loopColor.r = colorLoRGB.r * v_inverted + colorHiRGB.r * v;
            loopColor.g = colorLoRGB.g * v_inverted + colorHiRGB.g * v;
            loopColor.b = colorLoRGB.b * v_inverted + colorHiRGB.b * v;
            for (int i = 0; i < colors.Length; i++)
                colors[i] = loopColor;
            goTiles.GetComponent<MeshFilter>().mesh.colors = colors;
        }

        private void ColorLogic8() // "SSFX Trigger 1 Push [Hi/Lo Color]"
        {
            // framerate limit
            deltaTimeCounter += Time.deltaTime;
            if (deltaTimeCounter > 1f / 30f)
            {
                deltaTimeCounter = deltaTimeCounter - 1f / 30f;

                float v = avd.samples[avd.indexTrigger1];
                Color headColor = Color.HSVToRGB(jColorHi.val.H, jColorHi.val.S, jColorHi.val.V);
                headColor.a = jfAlpha.val;
                int snakeHeadPosition = (int)counter % (xSize * zSize) * 4;
                colors[snakeHeadPosition] = headColor;

                for (int i = 0; i < colors.Length; i += 4)
                {
                    // fade away color over time (* 0.95f), keep colorLoRGB minimum
                    if (colors[i].r <= colorLoRGB.r)
                        colors[i].r = colorLoRGB.r;
                    else
                        colors[i].r = colors[i].r * 0.95f;

                    if (colors[i].g <= colorLoRGB.g)
                        colors[i].g = colorLoRGB.g;
                    else
                        colors[i].g = colors[i].g * 0.95f;

                    if (colors[i].b <= colorLoRGB.b)
                        colors[i].b = colorLoRGB.b;
                    else
                        colors[i].b = colors[i].b * 0.95f;
                    colors[i + 1] = colors[i + 2] = colors[i + 3] = colors[i];
                }
                counter += v;
            }
            goTiles.GetComponent<MeshFilter>().mesh.colors = colors;
        }

        //////////////////////////////////////////////////////////////////////////////////
        // Geometry Logic
        //////////////////////////////////////////////////////////////////////////////////

        private void ResetMeshGeometry()
        {
            vertices = mesh.vertices;

            int i = 0;
            int x;
            for (int z = 0; z < zSize; z++)
            {
                for (x = 0; x < xSize; x++)
                {
                    // (x     , 0f, z + 1f);
                    vertices[i + 0].x = x;
                    vertices[i + 0].z = z + 1f;
                    // (x + 1f, 0f, z + 1f);
                    vertices[i + 1].x = x + 1f;
                    vertices[i + 1].z = z + 1f;
                    // (x     , 0f, z + 0f);
                    vertices[i + 2].x = x;
                    vertices[i + 2].z = z;
                    // (x + 1f, 0f, z + 0f);
                    vertices[i + 3].x = x + 1f;
                    vertices[i + 3].z = z;
                    i += 4;
                }
            }
            mesh.vertices = vertices;
        }

        private void GeometryLogicNone() { }

        private void GeometryLogicScaleTilesByBands()
        {
            vertices = mesh.vertices;

            int i = 0;
            int band = 0;
            int x;
            float xpos;
            float zpos;
            float v;
            for (int z = 0; z < zSize; z++)
            {
                zpos = z + 0.5f;
                for (x = 0; x < xSize; x++)
                {
                    xpos = x + 0.5f;
                    // fake maximum scale for testing
                    //v = halfTileScaleMin + 1f * (halfTileScaleMax - halfTileScaleMin);

                    // use min and max scale
                    v = halfTileScaleMin + avd.bands[band] * (halfTileScaleMax - halfTileScaleMin);
                    
                    // use max scale only
                    //v = avd.bands[band] * halfTileScaleMax;

                    // V1 Vector3(x     , 0f, z + 1f);
                    vertices[i    ].x = xpos - v;
                    //vertices[i + 0].y = 0f;
                    vertices[i    ].z = zpos + v;

                    // V2 Vector3(x + 1f, 0f, z + 1f);
                    vertices[i + 1].x = xpos + v;
                    //vertices[i + 1].y = 0f;
                    vertices[i + 1].z = vertices[i].z;

                    // V3 Vector3(x     , 0f, z + 0f);
                    vertices[i + 2].x = vertices[i].x;
                    //vertices[i + 2].y = 0;
                    vertices[i + 2].z = zpos - v;

                    // V4 Vector3(x + 1f, 0f, z + 0f);
                    vertices[i + 3].x = vertices[i + 1].x;
                    //vertices[i + 3].y = 0f;
                    vertices[i + 3].z = vertices[i + 2].z;

                    i += 4;
                    band++;
                    if (band >= avd.bands.Length)
                        band = 0;
                }
            }
            mesh.vertices = vertices;
        }

        private void GeometryLogicScaleTilesBySamples()
        {
            vertices = mesh.vertices;

            int i = 0;
            int sample = 0;
            int x;
            float xpos;
            float zpos;
            float v;
            for (int z = 0; z < zSize; z++)
            {
                zpos = z + 0.5f;
                for (x = 0; x < xSize; x++)
                {
                    xpos = x + 0.5f;
                    // fake maximum scale for testing
                    //v = halfTileScaleMin + 1f * (halfTileScaleMax - halfTileScaleMin);

                    // use min and max scale
                    v = halfTileScaleMin + avd.samples[sample] * (halfTileScaleMax - halfTileScaleMin);

                    // use max scale only
                    //v = avd.bands[band] * halfTileScaleMax;

                    // V1 Vector3(x     , 0f, z + 1f);
                    vertices[i].x = xpos - v;
                    //vertices[i + 0].y = 0f;
                    vertices[i].z = zpos + v;

                    // V2 Vector3(x + 1f, 0f, z + 1f);
                    vertices[i + 1].x = xpos + v;
                    //vertices[i + 1].y = 0f;
                    vertices[i + 1].z = vertices[i].z;

                    // V3 Vector3(x     , 0f, z + 0f);
                    vertices[i + 2].x = vertices[i].x;
                    //vertices[i + 2].y = 0;
                    vertices[i + 2].z = zpos - v;

                    // V4 Vector3(x + 1f, 0f, z + 0f);
                    vertices[i + 3].x = vertices[i + 1].x;
                    //vertices[i + 3].y = 0f;
                    vertices[i + 3].z = vertices[i + 2].z;

                    i += 4;
                    sample++;
                    if (sample >= avd.samples.Length)
                        sample = 0;
                }
            }
            mesh.vertices = vertices;
        }

        //////////////////////////////////////////////////////////////////////////////////
        // AVD Synchronization - called by SendMessage from Master
        //////////////////////////////////////////////////////////////////////////////////
        public void Sync(float[] syncdata)
        {
            try
            {
                avd.UnpackSyncdata(syncdata);
            }
            catch (Exception e) { SuperController.LogError("Exception: " + e); }
        }
    }
}