﻿// ***********************************************************************************
// SallySoundFX
// ***********************************************************************************
// Created by Sally Whitemane on VAM Hub
// https://www.patreon.com/sallyw
// Licensed under CC BY-SA - https://creativecommons.org/licenses/by-sa/4.0/

using System.Collections.Generic;
using UnityEngine;

namespace Sally
{
    public static class AVSlave
    {
        private static MVRScript masterScript;
        private static readonly float FIND_MASTER_RETRY_DELAY = 2.0f;
        private static float findMasterTimer = 0.0f;

        /// <summary>
        /// Updates the List of Slaves in the Master Plugin.
        /// </summary>
        public static void UpdateSlaves() { masterScript.SendMessage("UpdateSlaves"); }

        /// <summary>
        /// Use in Init() of Audio Visualization Slave (AVSLAVE) Plugin in to find and notify SSFX about new AVSLAVE.
        /// The SSFX Master Plugin is identified by it's AVD.MASTER_PLUGIN_IDSTRING-string in the class ("AVMASTER").
        /// A SSFX Slave Plugin is identified by it's AVD.SLAVE_PLUGIN_IDSTRING-string in the class ("AVSLAVE").
        /// </summary>
        public static MVRScript GetMasterScript(MVRScript self)
        {
            if (masterScript != null)
                return masterScript;

            List<Atom> atoms = self.GetSceneAtoms();
            foreach (Atom atom in atoms)
            {
                //Debug.Log("Atom: " + atom.name);
                MVRScript[] scripts = atom.GetComponentsInChildren<MVRScript>();
                foreach (MVRScript script in scripts)
                {
                    //Debug.Log("has script: " + script + "  name: " + script.name);
                    if (script.name.Contains(AVD.MASTER_PLUGIN_IDSTRING))
                    {
                        masterScript = script;
                        masterScript.SendMessage("UpdateSlaves");
                        return masterScript;
                    }
                }
            }
            return null;
        }

        /// <summary>
        /// Check sync status and searches for Master Plugin every 2 seconds if async.
        /// </summary>
        /// <returns>bool true if synchronized or false otherwise</returns>
        public static bool IsSynchronized(MVRScript self)
        {
            if (masterScript == null)
            {
                //Debug.Log($"{self.name} desynchronized. Waiting to retry ...");
                if (findMasterTimer < FIND_MASTER_RETRY_DELAY)
                {
                    // wait a short time before showing an error in case user just reloads SallySoundFX
                    findMasterTimer += Time.deltaTime;
                    return false;
                }
                //Debug.Log($"{self.name} desynchronized. Searching for Master Plugin...");
                findMasterTimer = 0;
                masterScript = GetMasterScript(self);
                if (masterScript == null)
                {
                    //Debug.Log($"{self.name} cannot find SallySoundFX Plugin. Retry in {FIND_MASTER_RETRY_DELAY} seconds.");
                    return false;
                }
                //Debug.Log($"{self.name} ... found!");
            }
            return true;
        }

        // Get directory path where the plugin is located. Based on Alazi's & VAMDeluxe's method.
        public static string GetPluginPath(MVRScript self)
        {
            // name: plugin#1_Sally.AVSLAVE_Tiles
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            // id: plugin#1
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            // filename: Custom/Scripts/Sally/SallySoundFX/AVTiles.cslist
            string r = filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
            // return: Custom/Sciprts/Sally/SallySoundsFX
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
        }

        // Get path prefix of the package that contains our plugin.
        public static string GetPackagePath(MVRScript self)
        {
            string id = self.name.Substring(0, self.name.IndexOf('_'));
            string filename = self.manager.GetJSON()["plugins"][id].Value;
            int idx = filename.IndexOf(":/");
            if (idx >= 0)
                return filename.Substring(0, idx + 2);
            else
                return string.Empty;
        }
    }
}