﻿// ***********************************************************************************
// SallySoundFX
// ***********************************************************************************
// Created by Sally Whitemane on VAM Hub
// https://www.patreon.com/sallyw
// Licensed under CC BY-SA - https://creativecommons.org/licenses/by-sa/4.0/

using System;
using System.Collections.Generic;
using UnityEngine;
using UnityEngine.UI;
using SimpleJSON;
using MacGruber;

namespace Sally {

	public class SSFX_AVMASTER : MVRScript
	{
		// UI - TOP
		private JSONStorableString jsTitleText;
		private UIDynamicTextField titleText;
		// UI - LEFT
		private JSONStorableFloat jfAmplification;
		private UIDynamicSlider sliderAmplification;
		private JSONStorableFloat jfSmooth;
		// UI - LEFT - Trigger 1
		private JSONStorableFloat jfTrigger1Freq;
		private FloatTrigger fTrigger1;
		// UI - LEFT - Trigger 2
		private JSONStorableFloat jfTrigger2Freq;
		private FloatTrigger fTrigger2;
		// UI - LEFT - Auto Amplification
		private JSONStorableBool jbAutoAmplification;
		// UI - LEFT - Bass Damper
		private JSONStorableBool jbBassDamper;
		private JSONStorableString jsBenchmark;
		// UI - RIGHT
		private JSONStorableFloat jfPeakFallSpeed;
		private JSONStorableFloat jfPeakGravity;
		private JSONStorableBool jbZoomSpectrum;
		private JSONStorableBool jbTestWave;
		private JSONStorableBool jbMicrophone;
        private JSONStorableBool jbAudioSrc;
        // BENCHMARK keep namespace, can cause VAM compilation error
        private System.Diagnostics.Stopwatch stopWatch = new System.Diagnostics.Stopwatch();
		private double[] avgUpdateTimes = new double[80];
		private double avgUpdateTime = 0;
		// AVD - AUDIO VISUALIZATION DATA
		private AVD avd = new AVD();
		private float[] syncdata;
		// MISC
		private bool InitOK = false;
		private List<MVRScript> slaves = new List<MVRScript>();
		private float deltaTimeDrawSinusLUT = 0f;
		private UISpectrum uiSpectrum;
		private GameObject goSpectrum;
        private float deltaTimeMS;
        private float peakSpeed;
        // AUDIO
        private AudioSource audioSource;
        private const float SILENCE_THRESHOLD = 0.0055f;
		private const float PEAK_DROP_SPEED = 0.00025f;
		private float peakDropSpeed = PEAK_DROP_SPEED;
		private const float PEAK_GRAVITY = 2.5f;
		private float peakGravity = PEAK_GRAVITY;
		private float samples_sum = 0f;
		private float samples_avg = 0f;
		// AUDIO - SAMPLES
		private static readonly int SAMPLES = AVD.SAMPLES;
		private float[] samplesUnity = new float[AVD.SAMPLES_UNITY];
		private float[] samples = new float[SAMPLES];
		private float[] samples_peak = new float[SAMPLES];
		private float[] samples_peakspeed = new float[SAMPLES];
		private float[] samples_previous_frame = new float[SAMPLES];
		// AUDIO - BANDS
		private static readonly int BANDS = AVD.BANDS;
		private float[] bands = new float[BANDS];
		private float[] bands_peak = new float[BANDS];
		private float[] bands_peakspeed = new float[BANDS];

		public override void Init()
		{
			base.Init();
			try {
				// Check if we are in subscene
				Atom atom = containingAtom; 
				if (containingAtom.parentAtom != null)
				{
					if (containingAtom.parentAtom.type.Equals("SubScene"))
					{
						// in SubScene
						atom = atom.parentAtom;
                    }
				}

                // Allow SallySoundFX only on atoms with AudioSource
                AudioSource[] audioSources = atom.GetComponentsInChildren<AudioSource>();
                if (audioSources.Length == 0)
                {
                    SuperController.LogError($"on {atom.uid} must have a AudioSource! Audio from the embedded browser will not work.");
                    return;
                }
                else if (audioSources.Length > 1)
                {
                    SuperController.LogError($"on {atom.uid} has multiple AudioSources!\nWARNING: Plugin may not work correct. The first found AudioSource will be used.");
                }
                audioSource = audioSources[0];

                // Prevent Plugin being loaded multiple times
                if (GetMasterScriptCount() > 1)
                {
					SuperController.LogError($"SallySoundFX already loaded. Only 1 instance is needed and allowed to avoid high CPU usage.");
					InitOK = false;
					return;
				}

				// Update List with slave Plugin
				UpdateSlaves();
				Mic.SelectByID(0);

				//////////////////////////////////////////////////////////////////////////////////
				// TOP
				//////////////////////////////////////////////////////////////////////////////////
				
				jsTitleText = new JSONStorableString("SallySoundFX", "<size=40><b><color=#FF60E0>Sally</color><color=#D080F0>Sound</color><color=#A0E0FF>FX</color></b></size>   <color=#A0A0A0>V" + AVD.VERSION.ToString("0.00") + "   Triggers and Visualizations   -   patreon.com/sallyw</color>");
				titleText = CreateTextField(jsTitleText, false);
				titleText.height = 340f;
				titleText.GetComponentInChildren<Image>().color = Color.black;

				goSpectrum = new GameObject("Spectrum");
				goSpectrum.transform.SetParent(titleText.transform, false);
				uiSpectrum = goSpectrum.AddComponent<UISpectrum>();
				uiSpectrum.transform.position = new Vector2(-515, -730);
				uiSpectrum.sampleIndexTrigger1 = avd.indexTrigger1;
				uiSpectrum.sampleIndexTrigger2 = avd.indexTrigger2;

				CreateSpacer(false).height = 325;
				CreateSpacer(true).height = 325;

				//////////////////////////////////////////////////////////////////////////////////
				// LEFT
				//////////////////////////////////////////////////////////////////////////////////
				UIDynamicSlider slider;
				UIDynamicButton button;

				// Slider Amplification
				jfAmplification = new JSONStorableFloat("Amplification", 20f, 1f, 2000f);
				RegisterFloat(jfAmplification);
				sliderAmplification = CreateSlider(jfAmplification, false);
				sliderAmplification.GetComponentInChildren<Image>().color = UI.blueAutoBackground;

				// Slider Smoothing
				jfSmooth = new JSONStorableFloat("Smoothing", 0.70f, 0.0f, 1.0f);
				RegisterFloat(jfSmooth);
				CreateSlider(jfSmooth, false);
				
				// Slider Trigger 1
				jfTrigger1Freq = new JSONStorableFloat("Trigger 1 Frequency in Hz", AVD.SampleFreq() * (avd.indexTrigger1 + 1), AVD.SampleFreq(), AVD.SampleFreq() * AVD.SAMPLES * 0.5f);
				RegisterFloat(jfTrigger1Freq);
				slider = CreateSlider(jfTrigger1Freq, false);
				slider.labelText.color = AVD.trigger1Color;
				slider.quickButtonsEnabled = false;
				slider.GetComponent<LayoutElement>().minWidth = 1060;
				jfTrigger1Freq.setCallbackFunction = SetTrigger1Frequency;

				// Slider Trigger 2
				jfTrigger2Freq = new JSONStorableFloat("Trigger 2 Frequency in Hz", AVD.SampleFreq() * (avd.indexTrigger2 + 1), AVD.SampleFreq(), AVD.SampleFreq() * AVD.SAMPLES * 0.5f);
				RegisterFloat(jfTrigger2Freq);
				slider = CreateSlider(jfTrigger2Freq, false);
				slider.quickButtonsEnabled = false;
				slider.GetComponent<LayoutElement>().minWidth = 1060;
				jfTrigger2Freq.setCallbackFunction = SetTrigger2Frequency;
				
				// Button Trigger 1
				fTrigger1 = new FloatTrigger(this, "SSFX Trigger 1");
				button = CreateButton("Trigger 1 Actions", false);
				button.buttonText.color = AVD.trigger1Color;
				button.button.onClick.AddListener(fTrigger1.OpenPanel);

				// Button Trigger 2
				fTrigger2 = new FloatTrigger(this, "SSFX Trigger 2");
				button = CreateButton("Trigger 2 Actions", false);
				button.buttonText.color = AVD.trigger2Color;
				button.button.onClick.AddListener(fTrigger2.OpenPanel);

				// Toggle Auto Amplification
				jbAutoAmplification = new JSONStorableBool("Auto Amplification", true, SetAutoAmplification);
				RegisterBool(jbAutoAmplification);
				CreateToggle(jbAutoAmplification, false);
				
				// Toggle Bass Damper
				jbBassDamper = new JSONStorableBool("Bass Damper", true);
				RegisterBool(jbBassDamper);
				CreateToggle(jbBassDamper, false);
				
				// Benchmark
				jsBenchmark = new JSONStorableString("Benchmark", "no data");
				UIDynamicTextField tf = CreateTextField(jsBenchmark, false);
				LayoutElement le = tf.GetComponent<LayoutElement>();
				le.minHeight = 40.0f;
				le.preferredHeight = 40.0f;

				//////////////////////////////////////////////////////////////////////////////////
				// RIGHT
				//////////////////////////////////////////////////////////////////////////////////

				// Slider Peak drop speed
				jfPeakFallSpeed = new JSONStorableFloat("Peak drop speed", PEAK_DROP_SPEED, PEAK_DROP_SPEED - 0.0002f, PEAK_DROP_SPEED + 0.001f);
				RegisterFloat(jfPeakFallSpeed);
				slider = CreateSlider(jfPeakFallSpeed, true);
				slider.Configure(jfPeakFallSpeed.name, jfPeakFallSpeed.min, jfPeakFallSpeed.max, jfPeakFallSpeed.val, jfPeakFallSpeed.constrained, "F4", false, false);
				jfPeakFallSpeed.setCallbackFunction = SetPeakFallSpeed;
				
				// Slider Peak gravity
				jfPeakGravity = new JSONStorableFloat("Peak gravity", PEAK_GRAVITY, 0f, 10f);
				RegisterFloat(jfPeakGravity);
				CreateSlider(jfPeakGravity, true);
				jfPeakGravity.setCallbackFunction = SetPeakGravity;
				
				// Spacer for wide sliders from left side
				CreateSpacer(true).height = 255;
				
				// Toggle Zoom
				jbZoomSpectrum = new JSONStorableBool("Zoom Spectrum to 64 samples", true);
				RegisterBool(jbZoomSpectrum);
				CreateToggle(jbZoomSpectrum, true);
                jbZoomSpectrum.setCallbackFunction = SetSpectrumSize;

                // Toggle Test Wave
                jbTestWave = new JSONStorableBool("Test wave if silent", false);
				RegisterBool(jbTestWave);
				CreateToggle(jbTestWave, true);
				
				// Toggle Microphone
				jbMicrophone = new JSONStorableBool("Microphone", false);
				RegisterBool(jbMicrophone);
				CreateToggle(jbMicrophone, true);
				jbMicrophone.setCallbackFunction = ToggleMicrophone;

                // AudioSource
                jbAudioSrc = new JSONStorableBool("AudioSrc", false);
                RegisterBool(jbAudioSrc);
                UIDynamicToggle toggle = CreateToggle(jbAudioSrc, true);
                toggle.labelText.text = "Listen to AudioSource only";

                // required for MacGruber_utils
                SimpleTriggerHandler.LoadAssets();
				InitOK = true;
			}
			catch (Exception e) { SuperController.LogError("Exception: " + e); }
		}
		public override void InitUI()
		{
			//Debug.Log("SallySoundFX InitUI()");
			base.InitUI();
			if (!InitOK)
				return;
			var scriptUI = UITransform.GetComponentInChildren<MVRScriptUI>();
			// breaks UI, stick with UIDynamicTextField for now
			//jsTitleText.text.transform.SetParent(scriptUI.fullWidthUIContent.transform, false);
			titleText.transform.SetParent(scriptUI.fullWidthUIContent.transform, false);
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				jc[fTrigger1.Name] = fTrigger1.GetJSON(base.subScenePrefix);
				jc[fTrigger2.Name] = fTrigger2.GetJSON(base.subScenePrefix);
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				fTrigger1.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				fTrigger2.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
			}
		}

		void Update()
		{
			if (!InitOK)
				return;
			try
			{
				// BENCHMARK START
				stopWatch.Reset();
				stopWatch.Start();

                deltaTimeMS = Time.deltaTime * 1000f;
                peakSpeed = 1f + Time.deltaTime * peakGravity;

                if (jbAudioSrc.val)
                    audioSource.GetSpectrumData(samplesUnity, 0, FFTWindow.Blackman);
                else
                    AudioListener.GetSpectrumData(samplesUnity, 0, FFTWindow.Blackman);
                Array.Copy(samplesUnity, samples, SAMPLES);

                // Analyze
                int i;
				samples_sum = 0f;
				for (i = 0; i < SAMPLES; i++)
					samples_sum += samples[i];
				samples_avg = samples_sum / SAMPLES;

				// Bass Damper
				if (jbBassDamper.val)
				{
					samples[0] *= 0.10f;
					samples[1] *= 0.10f;
					samples[2] *= 0.10f;
					samples[3] *= 0.12f;
					samples[4] *= 0.18f;
					samples[5] *= 0.25f;
					samples[6] *= 0.30f;
					samples[7] *= 0.35f;
					samples[8] *= 0.40f;
					samples[9] *= 0.45f;
					samples[10] *= 0.50f;
					samples[11] *= 0.55f;
					samples[12] *= 0.60f;
					samples[13] *= 0.65f;
					samples[14] *= 0.70f;
					samples[15] *= 0.75f;
					samples[16] *= 0.80f;
					samples[17] *= 0.85f;
					samples[18] *= 0.90f;
					samples[19] *= 0.95f;
				}

				// Amplify
				if (jbAutoAmplification.val)
					jfAmplification.val = (1.0f / samples_avg) * 0.33f;
				float amp = jfAmplification.val;
				for (i = 0; i < SAMPLES; i++)
				{
					samples[i] *= amp;
					if (samples[i] > 1f)
						samples[i] = 1f;
				}

				// Detect silence
				if (samples_sum < SILENCE_THRESHOLD)
				{
					if (jbTestWave.val)
					{
						// Create fake sinus test signal. Shift generated sinus LUT (look up table).
						// limit the array shift to 30 times per frames (or else it's to fast)
						deltaTimeDrawSinusLUT += Time.deltaTime;
						if (deltaTimeDrawSinusLUT > 1f / 30f)
						{
							deltaTimeDrawSinusLUT = deltaTimeDrawSinusLUT - 1f / 30f;
							float startToEnd = LUT.sin2x[0]; // store old start value
							Array.Copy(LUT.sin2x, 1, LUT.sin2x, 0, LUT.sin2x.Length - 1); // shift array
							LUT.sin2x[LUT.sin2x.Length - 1] = startToEnd; // move old start value to end
						}
                        Array.Copy(LUT.sin2x, samples, SAMPLES);
                    }
				}

				// Smoothing
				if (jfSmooth.val > 0.05f)
				{
					float smooth = jfSmooth.val;
					for (i = 0; i < SAMPLES; i++)
						samples[i] += (samples_previous_frame[i] - samples[i]) * smooth;
					Array.Copy(samples, samples_previous_frame, SAMPLES);
				}

				// Sample Peaks
				for (i = 0; i < SAMPLES; i++)
				{
					samples_peakspeed[i] *= peakSpeed;
					samples_peak[i] -= samples_peakspeed[i] * deltaTimeMS;
					if (samples_peak[i] < samples[i])
					{
						// new peak value, set peak to current value and reset peakSpeed
						samples_peak[i] = samples[i];
						samples_peakspeed[i] = peakDropSpeed;
					}
				}

				// Frequency Bands
				// band 0: 2  samples index 0-1
				// band 1: 2  samples index 2-3
				// band 2: 2  samples index 4-5
				// band 3: 4  samples index 6-9
				// band 4: 6  samples index 10-15
				// band 5: 16 samples index 16-31
				// band 6: 32 samples index 32-63
				// band 7: 64 samples index 64-127

				// band 0
				bands[0] = (samples[0] + samples[1]) * 0.5f;
				// band 1
				bands[1] = (samples[2] + samples[3]) * 0.5f;
				// band 2
				bands[2] = (samples[4] + samples[5]) * 0.5f;
				// band 3
				bands[3] = (samples[6] + samples[7] + samples[8] + samples[9]) * 0.25f;

                // band 4
                bands[4] = 0;
				for (i = 10; i < 16; i++)
					bands[4] += samples[i];
				bands[4] *= 0.16666666f;  // avoid slow divide by 6 => multiply by 0.166666f

				// band 5
				bands[5] = 0;
				for (i = 16; i < 32; i++)
					bands[5] += samples[i];
				bands[5] *= 0.0625f;   // avoid slow divide by 16 => multiply by 0.0625f

				// band 6
				bands[6] = 0;
				for (i = 32; i < 64; i++)
					bands[6] += samples[i];
				bands[6] *= 0.03125f;  // avoid slow divide by 32 => multiply by 0.03125f

				// band 7
				bands[7] = 0;
				for (i = 64; i < 127; i++)
					bands[7] += samples[i];
				bands[7] *= 0.015625f; // avoid slow divide by 64 => multiply by 0.015625f

				// Band peaks
				for (i = 0; i < BANDS; i++)
				{
					bands_peakspeed[i] *= peakSpeed;
                    bands_peak[i] -= bands_peakspeed[i] * deltaTimeMS;
					if (bands_peak[i] < bands[i])
					{
						// new peak value, set bands_peak to current value and reset bands_peakspeed
						bands_peak[i] = bands[i];
						bands_peakspeed[i] = peakDropSpeed;
					}
				}

				// Set sample values for Frequency Triggers based on user selected array index
				fTrigger1.Trigger(samples[avd.indexTrigger1]);
				fTrigger2.Trigger(samples[avd.indexTrigger2]);

				// Provide Audio Visualization Data for slaves
				// Make sure all slaves are valid
				foreach (MVRScript slave in slaves)
					if (slave == null) { UpdateSlaves(); break; }
				// Update AVD to pack it
				avd.bands = bands;
				avd.bands_peak = bands_peak;
				avd.samples = samples;
				avd.samples_peak = samples_peak;
				syncdata = avd.PackSyncdata();
				// Send AVD to slaves
				foreach (MVRScript slave in slaves)
					slave.SendMessage("Sync", syncdata);

				// Update Plugin UI Spectrum
				uiSpectrum.UpdateSamples(avd); //, bands_amp);

				// BENCHMARK END
				stopWatch.Stop();
				avgUpdateTimes[avgUpdateTimes.Length - 1] = stopWatch.Elapsed.TotalMilliseconds;
				Array.Copy(avgUpdateTimes, 1, avgUpdateTimes, 0, avgUpdateTimes.Length - 1); // shift array
				avgUpdateTime = 0;
				for (i = 0; i < avgUpdateTimes.Length; i++)
					avgUpdateTime += avgUpdateTimes[i];
				avgUpdateTime = avgUpdateTime / avgUpdateTimes.Length;

				jsBenchmark.val = $"Update Benchmark: {avgUpdateTime:F6} ms";
			}
			catch (Exception e)	{ SuperController.LogError("Exception: " + e); }
		}

		// FixedUpdate is called with each physics simulation frame by Unity
		//void FixedUpdate() {}

		// OnDestroy is where you should put any cleanup
		// if you registered objects to supercontroller or atom, you should unregister them here
		void OnDestroy()
		{
			if (fTrigger1 != null)
			{
				fTrigger1.Remove();
				fTrigger1 = null;
			}

			if (fTrigger2 != null)
			{
				fTrigger2.Remove();
				fTrigger2 = null;
			}

			avgUpdateTimes = null;
			avd = null;
			syncdata = null;

			// slaves
			if (slaves != null)
			{
				slaves.Clear();
				slaves = null;
			}

			uiSpectrum = null;

			// UI Spectrum
			if (goSpectrum != null)
			{
				Destroy(goSpectrum);
				goSpectrum = null;
			}

			samplesUnity = null;
			samples = null;
			samples_peak = null;
			samples_peakspeed = null;
			samples_previous_frame = null;

			bands = null;
			bands_peak = null;
			bands_peakspeed = null;
		}

		//////////////////////////////////////////////////////////////////////////////////
		// GUI Callbacks
		//////////////////////////////////////////////////////////////////////////////////
		
		private void SetAutoAmplification(bool b)
        {
			if (b)
				sliderAmplification.GetComponentInChildren<Image>().color = UI.blueAutoBackground;
			else
				sliderAmplification.GetComponentInChildren<Image>().color = UI.defaultBackground;
		}

		private void SetPeakFallSpeed(float v) {peakDropSpeed = v;}

		private void SetPeakGravity(float v) {peakGravity = v;}

		private void SetSpectrumSize(bool b)
		{
			if (b)
			{
				uiSpectrum.size = 64;
				jfTrigger1Freq.max = AVD.SampleFreq() * SAMPLES * 0.5f;
				jfTrigger2Freq.max = AVD.SampleFreq() * SAMPLES * 0.5f;
			}
			else
			{
				uiSpectrum.size = 128;
				jfTrigger1Freq.max = AVD.SampleFreq() * SAMPLES;
				jfTrigger2Freq.max = AVD.SampleFreq() * SAMPLES;
			}
		}

		private void SetTrigger1Frequency(float f)
		{
			f = AVD.SnapToSampleFrequency(f);
			jfTrigger1Freq.val = f;
			avd.indexTrigger1 = (int)(f / AVD.SampleFreq()) -1;
			uiSpectrum.sampleIndexTrigger1 = avd.indexTrigger1;
		}

		private void SetTrigger2Frequency(float f) {
			f = AVD.SnapToSampleFrequency(f);
			jfTrigger2Freq.val = f;
			avd.indexTrigger2 = (int)(f / AVD.SampleFreq()) - 1;
			uiSpectrum.sampleIndexTrigger2 = avd.indexTrigger2;
		}
		private void ToggleMicrophone(bool b) { if (b) Mic.On(); else Mic.Off(); }

		// called from init() and from Slaves Init() via SendMessage()
		private void UpdateSlaves()
		{
			try
			{
				//Debug.Log($"{name} UpdateSlaves()");
				slaves.Clear();
				List<Atom> atoms = GetSceneAtoms();
				foreach (Atom atom in atoms)
				{
					MVRScript[] scripts = atom.GetComponentsInChildren<MVRScript>();
					foreach (MVRScript script in scripts)
					{
						if (script.name.Contains(AVD.SLAVE_PLUGIN_IDSTRING))
						{
							// When this function is called from OnDestory() by a slave there is still a valid script, but the atom is already null
							if (script.containingAtom != null)
							{
								//Debug.Log($"{name} slaved.Add({script}) on Atom {script.containingAtom}");
								slaves.Add(script);
							}
						}
					}
				}
			}
			catch (Exception e) { SuperController.LogError("Exception: " + e); }
		}

		private int GetMasterScriptCount()
		{
			int count = 0;
			List<Atom> atoms = GetSceneAtoms();
			foreach (Atom atom in atoms)
			{
				//Debug.Log("Atom: " + atom.name);
				MVRScript[] scripts = atom.GetComponentsInChildren<MVRScript>();
				foreach (MVRScript script in scripts)
				{
					//Debug.Log("has script: " + script + "  name: " + script.name);
					if (script.name.Contains(AVD.MASTER_PLUGIN_IDSTRING))
						count++;
				}
			}
			return count;
		}
	}
}